"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_key_pair
hub.exec.boto3.client.ec2.delete_key_pair
hub.exec.boto3.client.ec2.describe_key_pairs
hub.exec.boto3.client.ec2.import_key_pair
resource = hub.tool.boto3.resource.crate(ctx,"ec2", "KeyPair", name)
"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    public_key: str,
    resource_id: str = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an ED25519 or 2048-bit RSA key pair with the specified name and
    in the specified PEM or PPK format that you can use with an Amazon EC2 instance.
    A key pair is used to control login access to EC2 instances. The key pair is available
    only in the Amazon Web Services Region in which you create it.
    Currently, this resource requires an existing user-supplied key pair.
    This key pair's public key will be registered with AWS to allow logging-in to EC2 instances.
    When importing an existing key pair the public key material may be in any format supported by AWS.
    Supported formats are described in Amazon docs:
    https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-key-pairs.html#how-to-generate-your-own-key-and-import-it-to-aws
    For more information, see the Amazon ec2 Developer Guide.
    Please note that, for security reasons, you cannot download the keypairs themselves.
    You'll simply be given their name and fingerprint.

    Args:
        name(str): The name for your new key pair.
        public_key(str): The public key material.
        resource_id(str, optional): The key pair id.
        tags(Dict, optional): dict of tags to assign to the key pair in the format of
        {"tag-key": "tag-value"} or dict in the format of {tag-key: tag-value}. Defaults to None.
            * tag-key (str) --  The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters.
            * tag-value (str) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.

    Request Syntax:
        [key_pair_name]:
          aws.ec2.key_pair.present:
          - resource_id: 'string'
          - name: 'string'
          - public_key: 'string'
          - tags: Dict[str, str]
              'string': 'string'
              'string': 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.ec2.key_pair.present:
                - name: value
                - public_key: value
                - tags:
                    first_key: first_value
                    second_key: second_value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.ec2.key_pair.get(
            ctx, name=name, filters=[{"name": "key-pair-id", "values": [resource_id]}]
        )

        if not before.get("result") and not before.get("ret"):
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        old_tags = result["old_state"].get("tags")

        if tags is not None and tags != old_tags:
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx,
                resource_id=resource_id,
                old_tags=old_tags,
                new_tags=tags,
            )
            result["result"] = update_ret["result"]
            result["comment"] = update_ret["comment"]

            if not update_ret["result"]:
                return result
            resource_updated = resource_updated or bool(update_ret["result"])

        if resource_updated:
            if ctx.get("test", False):
                plan_state["tags"] = update_ret["ret"]
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    "aws.ec2.key_pair", name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    "aws.ec2.key_pair", name
                )
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                "aws.ec2.key_pair", name
            )
    else:
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.key_pair", name=name
            )
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "public_key": public_key,
                    "tags": tags,
                },
            )
            return result

        ret = await hub.exec.boto3.client.ec2.import_key_pair(
            ctx,
            KeyName=name,
            PublicKeyMaterial=public_key,
            TagSpecifications=[
                {
                    "ResourceType": "key-pair",
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                }
            ]
            if tags
            else None,
        )

        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.key_pair", name=name
        )

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.ec2.key_pair.get(ctx, name=name)
        if not after["result"]:
            result["comment"] = after["comment"]
            result["result"] = after["result"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified key pair.

    Args:
        resource_id(str): An identifier of the resource in the provider.
        name(str, optional): Name of the key_pair. Defaults to None.

    Request Syntax:
        [key_pair_name]:
          aws.ec2.key_pair.absent:
          - resource_id: 'string'
          - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.key_pair.absent:
                - resource_id:(str): An identifier of the resource in the provider. Defaults to None.
                - name(str): value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.key_pair", name=name
        )
        return result

    before = await hub.exec.aws.ec2.key_pair.get(
        ctx, name=name, filters=[{"name": "key-pair-id", "values": [resource_id]}]
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.key_pair", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = copy.deepcopy(before["ret"])
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.key_pair", name=name
        )
        return result
    else:
        result["old_state"] = copy.deepcopy(before["ret"])
        ret = await hub.exec.boto3.client.ec2.delete_key_pair(ctx, KeyName=name)
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.key_pair", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns information about all key pairs in the user's account.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.key_pair
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_key_pairs(ctx, IncludePublicKey=True)

    if not ret["result"]:
        hub.log.debug(f"Could not describe key_pair {ret['comment']}")
        return result

    for key_pair in ret["ret"]["KeyPairs"]:
        resource_name = key_pair.get("KeyName")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_key_pair_to_present(
                raw_resource=key_pair
            )
        )
        result[resource_name] = {
            "aws.ec2.key_pair.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""State module for managing IAM Group Membership."""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub, ctx, name: str, group: str, users: List, resource_id: str = None
) -> Dict[str, Any]:
    """Adds the specified user to the specified group.

    Args:
        name(str): A name to represent the operation. This name is only for logging purpose. It is not used to
            attach a user to a group.

        group(str): Group Name

        users(List): List of users to attach to the group.

        resource_id(Text, Optional): The identifier for this object. Group Name

    Request Syntax:
        .. code-block:: sls

            [iam-group-membership]:
              aws.iam.group_membership.present:
                - name: 'string'
                - resource_id: 'string'
                - group: 'string'
                - users: list

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            iam-group-membership-temp-name:
              aws.iam.group_membership.present:
                - name: test_group
                - resource_id: test_group
                - group: test_group
                - users:
                  - test_user

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.iam.group_membership.get(
            ctx, name=group, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] += hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.iam.group_membership", name=name
        )

        update_ret = await hub.tool.aws.iam.group_membership.update(
            ctx,
            resource_id=resource_id,
            before=before["ret"],
            users=users,
        )
        result["comment"] += update_ret["comment"]
        if not update_ret["result"]:
            result["result"] = False
            return result

        resource_updated = bool(update_ret["ret"])

        if update_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_ret["ret"])

        if resource_updated and ctx.get("test", False):
            return result

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": group,
                    "group": group,
                    "users": users,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.iam.group_membership", name=group
            )
            return result

        for user in users:
            create_ret = await hub.exec.boto3.client.iam.add_user_to_group(
                ctx, GroupName=group, UserName=user
            )
            if not create_ret["result"]:
                result["result"] = False
                result["comment"] = create_ret["comment"]
                return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.iam.group_membership", name=group
        )
        resource_id = group

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.iam.group_membership.get(
            ctx, name=resource_id, resource_id=resource_id
        )
        if not after_ret["result"] or not after_ret["ret"]:
            result["result"] = False
            result["comment"] += tuple(after_ret["comment"])
            return result
        result["new_state"] = after_ret["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    users: List,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Removes users from the specified Group.

    Args:
        name(str): The name of the AWS IAM Group.
        users(List): List of users to remove.
        resource_id(str): Name of IAM group.

    Request Syntax:
        .. code-block:: sls

            [group_name]:
              aws.iam.group.absent:
                - name: "string"
                - resource_id: "string"
                - users:
                  - "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            iam-group-membership-test-group:
              aws.iam.group_membership.absent:
                - name: test_group
                - resource_id: test_group
                - users:
                  - test_user
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.group_membership", name=name
        )
        return result

    before = await hub.exec.aws.iam.group_membership.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.group_membership", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.iam.group_membership", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]

        for user in users:
            if not ctx.get("test", False):
                remove_usr_ret = await hub.exec.boto3.client.iam.remove_user_from_group(
                    ctx=ctx, GroupName=resource_id, UserName=user
                )
                if not remove_usr_ret["result"]:
                    result["comment"] = remove_usr_ret["comment"]
                    result["result"] = False
                    return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.iam.group_membership", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Gets information about the IAM Group membership.

    Lists the names of the users in IAM group.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.iam.group_membership

    """
    result = {}
    ret_groups = await hub.exec.boto3.client.iam.list_groups(ctx)
    if not ret_groups["result"]:
        result["comment"] = ret_groups["comment"]
        result["result"] = False
        return result

    group_name_list = [group.get("GroupName") for group in ret_groups["ret"]["Groups"]]

    for group_name in group_name_list:
        ret_group_users = await hub.exec.boto3.client.iam.get_group(
            ctx=ctx, GroupName=group_name
        )
        if not ret_group_users["result"]:
            result["comment"] = ret_group_users["comment"]
            result["result"] = False
            return result
        else:
            users = ret_group_users["ret"].get("Users")

            translated_resource = hub.tool.aws.iam.conversion_utils.convert_raw_group_membership_to_present(
                group=group_name, users=[user.get("UserName") for user in users]
            )

            resource_id = f"iam-group-membership-{group_name}"
            result[resource_id] = {
                "aws.iam.group_membership.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in translated_resource.items()
                ]
            }
    return result



"""Exec module for managing EC2 placement groups."""
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """Get an EC2 placement group resource from AWS. Supply one of the inputs as the filter.

    Get a placement group by either resource_id (GroupName) or filters.

    Args:
        name(str):
            The name of the Idem state and the GroupName of the placement group.
        resource_id(str, Optional):
            AWS placement group group name to identify the resource.
        filters(list, Optional):
            One or more filters: for example, ``tag:<key>``, ``tag-key``.
            A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_placement_groups

    Request Syntax:
        .. code-block:: bash

            $ idem exec aws.ec2.placement_group.get name=string resouce_id=string filters=list

    Returns:
        Dict

    Examples:
        .. code-block:: bash

            $ idem exec aws.ec2.placement_group.get name=idem_placement_group_name
            $ idem exec aws.ec2.placement_group.get name=null resource_id=my_placement_group
            $ idem exec aws.ec2.placement_group.get name=null filters="[{'name': 'group-name', 'values':['my_placement_group']}]"
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.tool.aws.ec2.placement_group.search_raw(
        ctx=ctx,
        name=name,
        filters=filters,
        resource_id=resource_id,
    )

    if not ret["result"]:
        if "InvalidPlacementGroup.Unknown" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.placement_group", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["PlacementGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.placement_group", name=name
            )
        )
        return result

    resource = ret["ret"]["PlacementGroups"][0]
    if len(ret["ret"]["PlacementGroups"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type="aws.ec2.placement_group", resource_id=resource_id
            )
        )
    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_placement_group_to_present(
        raw_resource=resource
    )
    return result


async def list_(hub, ctx, name, filters: List = None) -> Dict:
    """Get a list of EC2 placement group resources from AWS. Supply one of the inputs as the filter.

    Get a list of placement groups by idem name (group name) or filters.

    Args:
        name(str):
            The name of the Idem state and the GroupName of the placement group.
        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key.
            A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_placement_groups

    Request Syntax:
        .. code-block:: bash

            $ idem exec aws.ec2.placement_group.get name=string filters=list

    Returns:
        Dict

    Examples:
        .. code-block:: bash

            $ idem exec aws.ec2.placement_group.get name=idem_placement_group_name
            $ idem exec aws.ec2.placement_group.get name=null filters="[{'name': 'group-name', 'values':['my_placement_group']}]"
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.ec2.placement_group.search_raw(
        ctx=ctx,
        name=name,
        filters=filters,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["PlacementGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.placement_group", name=name
            )
        )
        return result
    for placement_group in ret["ret"]["PlacementGroups"]:
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_placement_group_to_present(
                raw_resource=placement_group
            )
        )
    return result



"""State module for managing AWS Budgets."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    budget_name: str,
    time_unit: str,
    budget_type: str,
    budget_limit: make_dataclass("Spend", [("Amount", str), ("Unit", str)]) = None,
    planned_budget_limits: Dict[
        str, make_dataclass("Spend", [("Amount", str), ("Unit", str)])
    ] = None,
    cost_filters: Dict[str, List[str]] = None,
    cost_types: make_dataclass(
        "CostTypes",
        [
            ("IncludeTax", bool, field(default=None)),
            ("IncludeSubscription", bool, field(default=None)),
            ("UseBlended", bool, field(default=None)),
            ("IncludeRefund", bool, field(default=None)),
            ("IncludeCredit", bool, field(default=None)),
            ("IncludeUpfront", bool, field(default=None)),
            ("IncludeRecurring", bool, field(default=None)),
            ("IncludeOtherSubscription", bool, field(default=None)),
            ("IncludeSupport", bool, field(default=None)),
            ("IncludeDiscount", bool, field(default=None)),
            ("UseAmortized", bool, field(default=None)),
        ],
    ) = None,
    time_period: make_dataclass(
        "TimePeriod",
        [
            ("Start", datetime, field(default=None)),
            ("End", datetime, field(default=None)),
        ],
    ) = None,
    auto_adjust_data: make_dataclass(
        "AutoAdjustData",
        [
            ("AutoAdjustType", str),
            (
                "HistoricalOptions",
                make_dataclass(
                    "HistoricalOptions",
                    [
                        ("BudgetAdjustmentPeriod", int),
                        ("LookBackAvailablePeriods", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("LastAutoAdjustTime", str, field(default=None)),
        ],
    ) = None,
    notifications_with_subscribers: List[
        make_dataclass(
            "NotificationWithSubscribers",
            [
                (
                    "Notification",
                    make_dataclass(
                        "Notification",
                        [
                            ("NotificationType", str),
                            ("ComparisonOperator", str),
                            ("Threshold", float),
                            ("ThresholdType", str, field(default=None)),
                            ("NotificationState", str, field(default=None)),
                        ],
                    ),
                ),
                (
                    "Subscribers",
                    List[
                        make_dataclass(
                            "Subscriber", [("SubscriptionType", str), ("Address", str)]
                        )
                    ],
                ),
            ],
        )
    ] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Creates an AWS Budget and, if included, notifications and subscribers.

    Args:
        name(str):
            An Idem name of the AWS Budget.
        resource_id(str):
            Resource ID to identify the AWS Budget.
        budget_name(str):
            The name of a budget. The name must be unique within an account. The ":" and "\\" characters aren't allowed
            in `budget_name`.
        time_unit(str):
            The length of time until a budget resets the actual and forecasted spend.
            'DAILY'|'MONTHLY'|'QUARTERLY'|'ANNUALLY'
        budget_type(str):
            Specifies whether this budget tracks costs, usage, RI utilization, RI coverage, Savings Plans utilization,
            or Savings Plans coverage.
        budget_limit(dict, Optional):
            The total amount of cost, usage, RI utilization, RI coverage, Savings Plans utilization, or Savings Plans
            coverage that you want to track with your budget.

            `budget_limit` is required for cost or usage budgets, but optional for RI or Savings Plans utilization
            or coverage budgets. RI and Savings Plans utilization or coverage budgets default to `100`. This is the
            only valid value for RI or Savings Plans utilization or coverage budgets. You can't use `budget_limit`
            with `planned_budget_limits` for `CreateBudget` and `UpdateBudget actions`.

            * Amount (str) -- [REQUIRED]
                The cost or usage amount that's associated with a budget forecast, actual spend, or budget threshold.

            * Unit (str) -- [REQUIRED]
                The unit of measurement that's used for the budget forecast, actual spend, or budget threshold,
                such as USD or GBP.
        planned_budget_limits(dict[str, Any], Optional):
            A map containing multiple `budget_limit`, including current or future limits.

            `planned_budget_limits` is available for cost or usage budget and supports both monthly and quarterly
            `time_unit`.

            For monthly budgets, provide 12 months of PlannedBudgetLimits values. This must start from the current
            month and include the next 11 months. The key is the start of the month, UTC in epoch seconds.

            For quarterly budgets, provide four quarters of PlannedBudgetLimits value entries in standard calendar
            quarter increments. This must start from the current quarter and include the next three quarters. The key
            is the start of the quarter, UTC in epoch seconds.

            If the planned budget expires before 12 months for monthly or four quarters for quarterly, provide the
            PlannedBudgetLimits values only for the remaining periods.

            If the budget begins at a date in the future, provide PlannedBudgetLimits values from the start date of the
            budget.

            After all the budget_limit values in planned_budget_limits are used, the budget continues to use the last
            limit as the budget_limit. At that point, the planned budget provides the same experience as a fixed budget.

            `DescribeBudget` and `DescribeBudgets` response along with `planned_budget_limits` also contain
            `budget_limit` representing the current month or quarter limit present in `planned_budget_limits`. This
            only applies to budgets that are created with `planned_budget_limits`. Budgets that are created without
            `planned_budget_limits` only contain `budget_limit`. They don't contain `planned_budget_limits`.

            The amount of cost or usage that's measured for a budget.

            For example, a Spend for 3 GB of S3 usage has the following parameters:
              * An `Amount` of `3`
              * A `unit` of `GB`

            * Amount (str):
                The cost or usage amount that's associated with a budget forecast, actual spend, or budget threshold.

            * Unit (str):
                The unit of measurement that's used for the budget forecast, actual spend, or budget threshold, such as
                USD or GBP.
        cost_filters(dict[str, list[str]], Optional):
            The cost filters, such as Region , Service , member account , Tag , or Cost Category , that are applied to
            a budget.

            Amazon Web Services Budgets supports the following services as a Service filter for RI budgets:
              * Amazon EC2
              * Amazon Redshift
              * Amazon Relational Database Service
              * Amazon ElastiCache
              * Amazon OpenSearch Service
        cost_types(dict[str, Any], Optional):
            The types of costs that are included in this COST budget.

            USAGE, RI_UTILIZATION, RI_COVERAGE, SAVINGS_PLANS_UTILIZATION, and SAVINGS_PLANS_COVERAGE budgets do
            not have `cost_types`.

            * IncludeTax (bool, Optional):
                Specifies whether a budget includes taxes. The default value is true.

            * IncludeSubscription (bool, Optional):
                Specifies whether a budget includes subscriptions. The default value is true.

            * UseBlended (bool, Optional):
                Specifies whether a budget uses a blended rate. The default value is false.

            * IncludeRefund (bool, Optional):
                Specifies whether a budget includes refunds. The default value is true.

            * IncludeCredit (bool, Optional):
                Specifies whether a budget includes credits. The default value is true.

            * IncludeUpfront (bool, Optional):
                Specifies whether a budget includes upfront RI costs. The default value is true.

            * IncludeRecurring (bool, Optional):
                Specifies whether a budget includes recurring fees such as monthly RI fees. The default value is true.

            * IncludeOtherSubscription (bool, Optional):
                Specifies whether a budget includes non-RI subscription costs. The default value is true.

            * IncludeSupport (bool, Optional):
                Specifies whether a budget includes support subscription fees. The default value is true.

            * IncludeDiscount (bool, Optional):
                Specifies whether a budget includes discounts. The default value is true.

            * UseAmortized (bool, Optional):
                Specifies whether a budget uses the amortized rate. The default value is false.
        time_period (dict[str, Any], Optional):
            The period of time that's covered by a budget. You set the start date and end date. The start
            date must come before the end date. The end date must come before `06/15/87 00:00 UTC`.

            If you create your budget and don't specify a start date, Amazon Web Services defaults to the start of
            your chosen time period (DAILY, MONTHLY, QUARTERLY, or ANNUALLY). For example, if you created
            your budget on January 24, 2018, chose `DAILY`, and didn't set a start date, Amazon Web Services
            set your start date to `01/24/18 00:00 UTC`. If you chose `MONTHLY`, Amazon Web Services set your
            start date to `01/01/18 00:00 UTC`. If you didn't specify an end date, Amazon Web Services set
            your end date to `06/15/87 00:00 UTC`. The defaults are the same for the Billing and Cost
            Management console and the API.

            You can change either date with the `UpdateBudget` operation.

            After the end date, Amazon Web Services deletes the budget and all the associated notifications
            and subscribers.

            * Start (datetime, Optional):
                The start date for a budget. If you created your budget and didn't specify a start date, Amazon
                Web Services defaults to the start of your chosen time period (DAILY, MONTHLY, QUARTERLY, or
                ANNUALLY). For example, if you created your budget on January 24, 2018, chose `DAILY`, and didn't
                set a start date, Amazon Web Services set your start date to `01/24/18 00:00 UTC`. If you chose
                `MONTHLY`, Amazon Web Services set your start date to `01/01/18 00:00 UTC`. The defaults are the
                same for the Billing and Cost Management console and the API.

                You can change your start date with the `UpdateBudget` operation.

            * End (datetime, Optional):
                The end date for a budget. If you didn't specify an end date, Amazon Web Services set your end date to
                `06/15/87 00:00 UTC`. The defaults are the same for the Billing and Cost Management console and the API.

                After the end date, Amazon Web Services deletes the budget and all the associated notifications and
                subscribers. You can change your end date with the `UpdateBudget` operation.
        auto_adjust_data (dict[str, Any], Optional):
            The parameters that determine the budget amount for an auto-adjusting budget.

            * AutoAdjustType (str):-- [REQUIRED]
                The string that defines whether your budget auto-adjusts based on historical or forecasted data.

            * HistoricalOptions (dict[str, Any], Optional): --
                The parameters that define or describe the historical data that your auto-adjusting budget is based on.

                * BudgetAdjustmentPeriod (int) -- [REQUIRED]
                    The number of budget periods included in the moving-average calculation that determines your
                    auto-adjusted budget amount.
                    The maximum value depends on the `time_unit` granularity of the budget:

                    * For the `DAILY` granularity, the maximum value is `60`.
                    * For the `MONTHLY` granularity, the maximum value is `12`.
                    * For the `QUARTERLY` granularity, the maximum value is `4`.
                    * For the `ANNUALLY` granularity, the maximum value is `1`.

                * LookBackAvailablePeriods (int) --
                    The integer that describes how many budget periods in your `BudgetAdjustmentPeriod` are included in
                    the calculation of your current `BudgetLimit`. If the first budget period in your
                    `BudgetAdjustmentPeriod` has no cost data, then that budget period isn’t included in the average
                    that determines your budget limit.

                    For example, if you set `BudgetAdjustmentPeriod` as `4` quarters, but your account had no cost data
                    in the first quarter, then only the last three quarters are included in the calculation. In this
                    scenario, `LookBackAvailablePeriods` returns `3`.

                    You can’t set your own LookBackAvailablePeriods. The value is automatically calculated from the
                    BudgetAdjustmentPeriod and your historical cost data.

            * LastAutoAdjustTime (datetime) --
                The last time that your budget was auto-adjusted.
        notifications_with_subscribers(list[dict[str, Any]], Optional):
            A notification that you want to associate with a budget. A budget can have up to five notifications, and
            each notification can have one SNS subscriber and up to 10 email subscribers. If you include notifications
            and subscribers in your CreateBudget call, Amazon Web Services creates the notifications and subscribers
            for you.

            A notification with subscribers can have one SNS subscriber and up to 10 email subscribers, for a total of
            11 subscribers.

            * Notification (dict[str, Any]): -- [REQUIRED]
                The notification that's associated with a budget.

                * NotificationType (str) -- [REQUIRED]
                    Specifies whether the notification is for how much you have spent (ACTUAL) or for how much that
                    you're forecasted to spend (FORECASTED).

                * ComparisonOperator (str) -- [REQUIRED]
                    The comparison that's used for this notification.

                * Threshold (float) -- [REQUIRED]
                    The threshold that's associated with a notification. Thresholds are always a percentage, and many
                    customers find value being alerted between 50% - 200% of the budgeted amount. The maximum limit for
                    your threshold is 1,000,000% above the budgeted amount.

                * ThresholdType (str) --
                    The type of threshold for a notification. For `ABSOLUTE_VALUE` thresholds, Amazon Web Services
                    notifies you when you go over or are forecasted to go over your total cost threshold. For
                    `PERCENTAGE` thresholds, Amazon Web Services notifies you when you go over or are forecasted to go
                    over a certain percentage of your forecasted spend. For example, if you have a budget for 200
                    dollars, and you have a `PERCENTAGE` threshold of 80%, Amazon Web Services notifies you when you go
                    over 160 dollars.

                * NotificationState (str) --
                    Specifies whether this notification is in alarm. If a budget notification is in the `ALARM` state,
                    you passed the set threshold for the budget.

            * Subscribers (list[dict[str, Any]]): -- [REQUIRED]
                A list of subscribers who are subscribed to this notification.

                The subscriber to a budget notification consists of a subscription type and either an Amazon SNS topic
                or an email address.

                For example, an email subscriber has the following parameters:
                  * A `subscriptionType` of `EMAIL`
                  * An `address` of `example@example.com`

                * SubscriptionType (str) -- [REQUIRED]
                    The type of notification that Amazon Web Services sends to a subscriber.

                * Address (str) -- [REQUIRED]
                    The address that Amazon Web Services sends budget notifications to, either an SNS topic or an email.

                    When you create a subscriber, the value of `Address` can't contain line breaks.


    Request Syntax:
      .. code-block:: sls

        [budget-resource-id]:
          aws.budgets.budget.present:
            - resource_id: "string"
            - budget_name: "string"
            - time_unit: "string"
            - budget_type: "string"
            - budget_limit: "dict"
            - cost_filters: "dict"
            - cost_types: "dict"
            - time_period: "dict"
            - auto_adjust_data: "dict"
            - notifications_with_subscribers: "dict"

    Returns:
        Dict[str, str]

    Examples:
        .. code-block:: sls

            cost_budget1234:
              aws.budgets.budget.present:
                - budget_name: new-test-budget
                - budget_limit:
                    Amount: "700.0"
                    Unit: USD
                - cost_filters: {}
                - cost_types:
                    IncludeCredit: false
                    IncludeDiscount: true
                    IncludeOtherSubscription: true
                    IncludeRecurring: true
                    IncludeRefund: false
                    IncludeSubscription: true
                    IncludeSupport: true
                    IncludeTax: true
                    IncludeUpfront: true
                    UseAmortized: false
                    UseBlended: false
                - time_unit: MONTHLY
                - time_period:
                    End: 2087-06-15 05:30:00+05:30
                  Start: 2022-06-01 05:30:00+05:30
              - budget_type: COST

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]
    resource_updated = False

    budget_parameters = {
        "BudgetName": budget_name,
        "CostFilters": cost_filters,
        "CostTypes": cost_types,
        "TimeUnit": time_unit,
        "TimePeriod": time_period,
        "BudgetType": budget_type,
    }

    if budget_limit:
        budget_parameters["BudgetLimit"] = budget_limit
    else:
        budget_parameters["PlannedBudgetLimits"] = planned_budget_limits
    if auto_adjust_data:
        budget_parameters["AutoAdjustData"] = auto_adjust_data

    if resource_id:
        before = await hub.exec.boto3.client.budgets.describe_budget(
            ctx, AccountId=account_id, BudgetName=resource_id
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

    if before:
        result[
            "old_state"
        ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
            ctx,
            account_id=account_id,
            raw_resource=before["ret"]["Budget"],
            idem_resource_name=name,
        )
        plan_state = copy.deepcopy(result["old_state"])

        update_ret = await hub.tool.aws.budgets.budget.update_budget(
            ctx,
            account_id=account_id,
            before=before["ret"]["Budget"],
            budget_limit=budget_limit,
            planned_budget_limits=planned_budget_limits,
            cost_filters=cost_filters,
            cost_types=cost_types,
            time_unit=time_unit,
            time_period=time_period,
            budget_type=budget_type,
            auto_adjust_data=auto_adjust_data,
        )
        result["comment"] += update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            plan_state = update_ret["ret"]

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.budgets.budget", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.budgets.budget", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "budget_name": budget_name,
                    "budget_limit": budget_limit,
                    "cost_filters": cost_filters,
                    "cost_types": cost_types,
                    "time_unit": time_unit,
                    "time_period": time_period,
                    "budget_type": budget_type,
                    "auto_adjust_data": auto_adjust_data,
                    "notifications_with_subscribers": notifications_with_subscribers,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.budgets.budget", name=name
            )
            return result
        ret = await hub.exec.boto3.client.budgets.create_budget(
            ctx,
            AccountId=account_id,
            Budget=budget_parameters,
            NotificationsWithSubscribers=notifications_with_subscribers,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.budgets.budget", name=name
        )

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.budgets.describe_budget(
                ctx, AccountId=account_id, BudgetName=budget_name
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result[
                "new_state"
            ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
                ctx,
                account_id=account_id,
                raw_resource=after["ret"]["Budget"],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes an AWS Budget.

    Args:
        name(str): An Idem name of the AWS Budget.
        resource_id(str, Optional): Budget ID to identify the resource.
            Idem automatically considers this resource being absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [budget-resource-id]:
              aws.budgets.budget.absent:
                - name: value
                - resource_id: value

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.budgets.budget.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.budgets.budget", name=name
        )
        return result

    before = await hub.exec.boto3.client.budgets.describe_budget(
        ctx, AccountId=account_id, BudgetName=name
    )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.budgets.budget", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
            ctx,
            account_id=account_id,
            raw_resource=before["ret"]["Budget"],
            idem_resource_name=name,
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.budgets.budget", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
            ctx,
            account_id=account_id,
            raw_resource=before["ret"]["Budget"],
            idem_resource_name=name,
        )
        ret = await hub.exec.boto3.client.budgets.delete_budget(
            ctx, AccountId=account_id, BudgetName=name
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.budgets.budget", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Lists the AWS Budgets that are associated with an account.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.budgets.budget
    """
    result = {}
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]
    ret = await hub.exec.boto3.client.budgets.describe_budgets(
        ctx, AccountId=account_id
    )

    if not ret["result"]:
        hub.log.debug(f"Could not describe budgets {ret['comment']}")
        return {}

    for budget in ret["ret"]["Budgets"]:
        resource_id = budget.get("BudgetName")
        resource_translated = (
            await hub.tool.aws.budgets.conversion_utils.convert_raw_budget_to_present(
                ctx,
                account_id=account_id,
                raw_resource=budget,
                idem_resource_name=resource_id,
            )
        )
        result[resource_id] = {
            "aws.budgets.budget.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from typing import Any
from typing import Dict


def sig_gather(hub, profiles) -> Dict[str, Any]:
    ...


async def post_gather(hub, ctx) -> Dict[str, Any]:
    """
    This function is called after any "idem_aws.acct.aws.*.gather" function.
    Modifications to profiles should happen in plugins located in "idem_aws.tool.aws.acct".
    Modifications to profiles should NOT occur explicitly in this function.
    """
    profiles = ctx.ret or {}

    hub.log.info(f"Read {len(profiles)} profiles from {ctx.ref}")

    for name, raw_profile in profiles.items():
        new_profile = {}

        # Apply modifications to the profile before using it
        await hub.tool.aws.acct.init.modify(name, raw_profile, new_profile)

        # Log whether this is a new profile or if the mods are overwriting an existing profile
        if name in profiles:
            hub.log.debug(f"Modified profile '{name}' for the current run")
        else:
            hub.log.debug(f"Created new profile '{name}'")

        # Add the fully modified profile to all the profiles
        profiles[name] = new_profile

    return profiles



"""
hub.exec.boto3.client.ec2.create_flow_logs
hub.exec.boto3.client.ec2.delete_flow_logs
hub.exec.boto3.client.ec2.describe_flow_logs
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": ["aws.ec2.subnet.present"],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    resource_type: str,
    traffic_type: str,
    log_format: str = None,
    max_aggregation_interval: int = None,
    resource_id: str = None,
    log_destination: str = None,
    log_destination_type: str = None,
    log_group_name: str = None,
    iam_role: str = None,
    destination_options: make_dataclass(
        "DestinationOptionsRequest",
        [
            ("FileFormat", str, field(default=None)),
            ("HiveCompatiblePartitions", bool, field(default=None)),
            ("PerHourPartition", bool, field(default=None)),
        ],
    ) = None,
    resource_ids: List[str] = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""Creates a flow log to capture information about IP traffic for a specific network interface, subnet,
    or VPC.

    Flow log data for a monitored network interface is recorded as flow log records, which are log events
    consisting of fields that describe the traffic flow. For more information, see Flow log records in the Amazon
    Virtual Private Cloud User Guide. When publishing to CloudWatch Logs, flow log records are published to a log
    group, and each network interface has a unique log stream in the log group. When publishing to Amazon S3, flow
    log records for all of the monitored network interfaces are published to a single log file object that is stored
    in the specified bucket. For more information, see VPC Flow Logs in the Amazon Virtual Private Cloud User Guide.

    Args:
        name(str):
            Name of the resource.

        log_group_name(str, Optional):
            log_group_name if the log_destination_type is cloudwatch.

        resource_type(str):
            Type of resource flow-log is attached to (Subnet, VPC, NetworkInterface).

        traffic_type(str):
            Type of traffic to be recorded (REJECT, ALL, ACCEPT).

        log_destination_type(str, Optional):
            S3 bucket or Default: cloud-watch-logs.

        log_destination(str, Optional):
            S3 bucket ARN.

        log_format(str, Optional):
            Syntax to be used to print log statements.

        max_aggregation_interval(int, Optional):
            Max interval during which packets are aggregated and then stored in log record.

        resource_id(str, Optional):
            AWS Flow log ID.

        iam_role(str, Optional):
            ARN of IAM role to be used to post in cloud-watch-logs.

        destination_options(dict[str, Any], Optional):
            The destination options. Defaults to None.

            * FileFormat (str, Optional):
                The format for the flow log. The default is plain-text.

            * HiveCompatiblePartitions (bool, Optional):
                Indicates whether to use Hive-compatible prefixes for flow logs stored in Amazon S3. The default is
                false.

            * PerHourPartition (bool, Optional):
                Indicates whether to partition the flow log per hour. This reduces the cost and response time for
                queries. The default is false.

        resource_ids([str]):
            list of resource_ids flow-log is attached to.

        tags(dict or list, Optional):
            Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the flow log resource. Defaults to None.

            * (Key, Optional):
                The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.

            * (Value, Optional):
                The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        .. code-block:: sls

            [flow-log-id]:
              aws.ec2.flow_log.present:
                - name: 'string'
                - log_group_name: 'string'
                - resource_type: 'integer'
                - traffic_type: 'string'
                - tags:
                    - Key: 'string'
                      Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            fl-09c0787e693332a0a:
                aws.ec2.flow_log.present:
                  - traffic_type: REJECT
                  - log_destination_type: s3
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.ec2.flow_log.get(
            ctx, name=name, resource_id=resource_id
        )
    if isinstance(tags, List):
        tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    else:
        tags_dict = tags
    if before and before["result"] and before["ret"]:
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        old_tags = before["ret"].get("tags")
        if tags_dict is not None:
            if not hub.tool.aws.state_comparison_utils.compare_dicts(
                tags_dict, old_tags
            ):
                result["comment"] = f"'{name}' already exists. Updating tags"
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx,
                    resource_id=resource_id,
                    old_tags=old_tags,
                    new_tags=tags_dict,
                )
                if not update_ret["result"]:
                    result["comment"] = update_ret["comment"]
                    result["result"] = False
                if ctx.get("test", False) and update_ret["ret"] is not None:
                    plan_state["tags"] = update_ret["ret"]
                resource_updated = resource_updated or bool(update_ret["result"])
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_type": resource_type,
                    "traffic_type": traffic_type,
                    "log_format": log_format,
                    "max_aggregation_interval": max_aggregation_interval,
                    "resource_id": resource_id,
                    "log_destination": log_destination,
                    "log_destination_type": log_destination_type,
                    "log_group_name": log_group_name,
                    "iam_role": iam_role,
                    "destination_options": destination_options,
                    "resource_ids": resource_ids,
                    "tags": tags_dict,
                },
            )
            result["comment"] = f"Would create aws.ec2.flow_log {name}"
            return result
        try:
            if isinstance(tags, Dict):
                tags_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            else:
                tags_list = tags
            ret = await hub.exec.boto3.client.ec2.create_flow_logs(
                ctx,
                DeliverLogsPermissionArn=iam_role,
                TagSpecifications=[{"ResourceType": "vpc-flow-log", "Tags": tags_list}]
                if tags_list is not None
                else None,
                DestinationOptions=destination_options,
                LogDestination=log_destination,
                LogDestinationType=log_destination_type,
                LogFormat=log_format,
                LogGroupName=log_group_name,
                MaxAggregationInterval=max_aggregation_interval,
                ResourceIds=resource_ids,
                ResourceType=resource_type,
                TrafficType=traffic_type,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Created '{name}'"
            resource_id = ret["ret"]["FlowLogIds"][0]
            resource_updated = True
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif resource_updated:
        after = await hub.exec.aws.ec2.flow_log.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if after and after["result"] and after["ret"]:
            result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Deletes a flow log.

    Args:
        name(str): Name of the resource.

        resource_id(str, Optional): AWS Flow log ID.

    Request Syntax:
        .. code-block:: sls

            [flow-log-name]:
              aws.ec2.flow_log.present:
                - name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.ec2.flow_log.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.flow_log", name=name
        )
        return result

    before = await hub.exec.aws.ec2.flow_log.get(
        ctx=ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        if "InvalidFlowLogId.NotFound" in str(before["comment"]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.ec2.flow_log", name=name
            )
        else:
            result["result"] = False
            result["comment"] = before["comment"]
        return result
    elif not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.flow_log", name=name
        )
        return result

    result["old_state"] = before["ret"]
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.flow_log", name=name
        )
    else:
        ret = await hub.exec.boto3.client.ec2.delete_flow_logs(
            ctx, FlowLogIds=[resource_id]
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.flow_log", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Describes one or more flow logs. To view the information in your flow logs (the log streams for the network
    interfaces), you must use the CloudWatch Logs console or the CloudWatch Logs API.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws_auto.ec2.flow_log
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_flow_logs(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe flow_log {ret['comment']}")
        return result
    for flow_log in ret["ret"]["FlowLogs"]:
        translated_resource = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_flow_log_to_present(flow_log)
        )

        result[flow_log["FlowLogId"]] = {
            "aws.ec2.flow_log.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result



from typing import Dict
from typing import List


async def search_raw(
    hub, ctx, resource_id: str = None, filters: List = None, tags: List = None
) -> Dict:
    """
    Fetch one or more volumes from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        resource_id(string, optional): volume ID of the volume.
        filters(List, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_volumes
        tags(List, optional): The list of tags to filter by. For example, to find all resources that have a tag with the key
            "Owner" and the value "TeamA" , specify "tag:Owner" for the Dict key and "TeamA" for the Dict value.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    if tags is not None:
        for tag in tags:
            boto3_filter.append({"Name": tag["Key"], "Values": [tag["Value"]]})

    ret = await hub.exec.boto3.client.ec2.describe_volumes(
        ctx, Filters=boto3_filter, VolumeIds=[resource_id] if resource_id else None
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result



from typing import Dict
from typing import List


async def search_raw(
    hub,
    ctx,
    name,
    filters: List = None,
    resource_id: str = None,
) -> Dict:
    """
    Fetch one or more placement groups from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        name(string): The name of the Idem state and the GroupName of the placement group
        filters(list, optional): One or more filters. For example: {"Name": "strategy", "Values": ["cluster"]},
        resource_id(string, optional): The GroupName of the placement group.
        A complete list of filters can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_placement_groups
    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )

    ret = await hub.exec.boto3.client.ec2.describe_placement_groups(
        ctx,
        Filters=boto3_filter,
        GroupNames=[resource_id] if resource_id else None,
    )

    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.s3.bucket.present",
        ],
    },
}

BUCKET_NOT_EXISTS_ERROR_MESSAGE = "ClientError: An error occurred (404) when calling the HeadBucket operation: Not Found"


async def present(
    hub,
    ctx,
    name: str,
    bucket: str,
    policy: str,
    confirm_remove_self_bucket_access: bool = None,
    expected_bucket_owner: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Applies an Amazon S3 bucket policy to an Amazon S3 bucket. If you are using an identity other than the root user of
    the Amazon Web Services account that owns the bucket, the calling identity must have the PutBucketPolicy permissions
    on the specified bucket and belong to the bucket owner's account in order to use this operation. If you don't have
    PutBucketPolicy permissions, Amazon S3 returns a 403 Access Denied error. If you have the correct permissions, but
    you're not using an identity that belongs to the bucket owner's account, Amazon S3 returns a 405 Method Not Allowed error.

    Args:
        name(Text): The name of the bucket policy.
        bucket(Text):  The name of the S3 bucket
        resource_id(Text, optional): S3 Bucket policy ID
        policy(Text): The bucket policy as a JSON document.
        confirm_remove_self_bucket_access(boolean, optional): Set this parameter to true to confirm that you want to remove
                                                            your permissions to change this bucket policy in the future.
        expected_bucket_owner(Text, optional): The account ID of the expected bucket owner. If the bucket is owned
                                                by a different account, the request will fail with an HTTP 403 (Access Denied) error.

    Request Syntax:
        [bucket_name]-policy:
          aws.s3.bucket_policy.present:
            - bucket: string
            - policy: string
            - confirm_remove_self_bucket_access: boolean
            - expected_bucket_owner: string

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-s3-bucket-bb7bb32e9533-policy:
              aws.s3.bucket_policy.present:
                - bucket: my-s3-bucket-bb7bb32e9533
                - policy: '{"Version":"2012-10-17","Statement":[{"Sid":"PublicReadGetObject","Effect":"Allow","Principal":{"AWS":"arn:aws:iam::418235808912:root"},"Action":"s3:GetObject","Resource":"arn:aws:s3:::thebugbucket3/*"}]}'
                - confirm_remove_self_bucket_access: False
                - expected_bucket_owner: 1239234249


    """
    result = dict(comment=(), name=name, old_state=None, new_state=None, result=True)

    # Check if Policy state name has correct format i.e. [bucket_name]-policy
    if name.split("-policy")[0] != bucket:
        result["comment"] = (
            f"Incorrect aws.s3.bucket_policy name: {name}. Expected name {bucket}-policy",
        )
        result["result"] = False
        return result

    # Check if bucket exists
    ret = await hub.exec.boto3.client.s3.head_bucket(ctx, Bucket=bucket)
    if not ret["result"]:
        if BUCKET_NOT_EXISTS_ERROR_MESSAGE in ret["comment"]:
            result["comment"] = (f"aws.s3.bucket with name: {bucket} does not exist",)
        else:
            result["comment"] = ret["comment"]
        result["result"] = False
        return result

    before = None
    resource_updated = False
    plan_state = None

    # Get current bucket policy for the bucket
    if resource_id:
        before = await hub.exec.boto3.client.s3.get_bucket_policy(
            ctx, Bucket=bucket, ExpectedBucketOwner=expected_bucket_owner
        )
        if before and not before["result"]:
            result["comment"] = (
                f"Failed to get existing Bucket policy '{name}'. Error : "
                + before["comment"]
            )
            result["result"] = before["result"]
            return result

        result[
            "old_state"
        ] = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_policy_to_present(
            raw_resource=before["ret"], bucket=bucket, name=name
        )
        plan_state = copy.deepcopy(result["old_state"])
        result["comment"] = (f"Bucket policy '{name}' already present.",)
        if result["old_state"]["policy"] != policy:
            resource_updated = True
            plan_state["policy"] = policy

    # Handling test scenario seperately
    if ctx.get("test", False):
        if not before:
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "policy": policy,
                    "bucket": bucket,
                    "resource_id": bucket + "-policy",
                },
            )
            result["comment"] = (f"Would create aws.s3.bucket_policy '{name}'",)
        else:
            result["new_state"] = plan_state
            if resource_updated:
                result["comment"] = (f"Would update aws.s3.bucket_policy '{name}'",)
            else:
                result["comment"] = (f"No change in policy.",)
        return result

    # If policy is not present or if it is updated
    if (not before) or resource_updated:
        ret = await hub.exec.boto3.client.s3.put_bucket_policy(
            ctx,
            Bucket=bucket,
            Policy=policy,
            ConfirmRemoveSelfBucketAccess=confirm_remove_self_bucket_access,
            ExpectedBucketOwner=expected_bucket_owner,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        if resource_updated:
            result["comment"] = result["comment"] + (f"Policy updated successfully",)
        else:
            result["comment"] = (f"Bucket policy '{name}' created.",)
        after = await hub.exec.boto3.client.s3.get_bucket_policy(
            ctx, Bucket=bucket, ExpectedBucketOwner=expected_bucket_owner
        )
        result[
            "new_state"
        ] = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_policy_to_present(
            raw_resource=after["ret"], bucket=bucket, name=name
        )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = result["comment"] + (f"No change in policy.",)
    return result


async def absent(
    hub,
    ctx,
    name: str,
    bucket: str,
    resource_id: str = None,
    expected_bucket_owner: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the policy of specified s3 bucket.

    Args:
        name(Text): The name of the bucket policy.
        bucket(Text):  The name of the S3 bucket
        resource_id(Text, optional): S3 Bucket policy ID. Idem automatically considers this resource being absent if this field is not specified.
        expected_bucket_owner(Text, optional): The account ID of the expected bucket owner. If the bucket is owned by a
                                    different account, the request will fail with an HTTP 403 (Access Denied) error.

    Returns:
        Dict[str, Any]

    Request Syntax:
        [bucket_name]-policy:
          aws.s3.bucket_policy.absent:
            - bucket: string
            - expected_bucket_owner: string

    Examples:

        .. code-block:: sls

            my-s3-bucket-bb7bb32e9533-policy:
              aws.s3.bucket_policy.absent:
                - bucket: my-s3-bucket-bb7bb32e9533
                - expected_bucket_owner: 1239234249
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket_policy", name=name
        )
        return result

    if resource_id.split("-policy")[0] != bucket:
        result["comment"] = (f"Incorrect aws.s3.bucket name: '{bucket}'",)
        result["result"] = False
        return result

    # Check if bucket exists.
    ret = await hub.exec.boto3.client.s3.head_bucket(ctx, Bucket=bucket)
    if not ret["result"]:
        if BUCKET_NOT_EXISTS_ERROR_MESSAGE in ret["comment"]:
            result["comment"] = (f"aws.s3.bucket with name: '{bucket}' does not exist",)
        else:
            result["comment"] = ret["comment"]
        result["result"] = False
        return result

    before = await hub.exec.boto3.client.s3.get_bucket_policy(
        ctx, Bucket=bucket, ExpectedBucketOwner=expected_bucket_owner
    )

    if before and before["result"]:
        result[
            "old_state"
        ] = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_policy_to_present(
            raw_resource=before["ret"], bucket=bucket, name=name
        )

        if ctx.get("test", False):
            result["comment"] = (f"Would delete aws.s3.bucket_policy '{name}'",)
            return result

        ret = await hub.exec.boto3.client.s3.delete_bucket_policy(
            ctx, Bucket=bucket, ExpectedBucketOwner=expected_bucket_owner
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = (f"Deleted policy '{name}' for bucket '{bucket}'",)
        return result
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket_policy", name=name
        )
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Obtain S3 bucket policy for each bucket under the given context for any user.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.s3.bucket_policy
    """
    result = {}
    # To describe the policy of all the buckets, we first need to list all the buckets, then get the
    # policy of each bucket
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe S3 buckets {ret['comment']}")
        return {}

    for bucket in ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")
        # get policy for each bucket
        bucket_policy = await hub.exec.boto3.client.s3.get_bucket_policy(
            ctx, Bucket=bucket_name
        )
        if bucket_policy["result"]:
            bucket_policy_name = f"{bucket_name}-policy"
            # translate policy to a common format
            policy_translated = (
                hub.tool.aws.s3.conversion_utils.convert_raw_bucket_policy_to_present(
                    raw_resource=bucket_policy["ret"],
                    bucket=bucket_name,
                    name=bucket_policy_name,
                )
            )
            result[bucket_policy_name] = {
                "aws.s3.bucket_policy.present": [policy_translated]
            }
    return result



from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str = None,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """
    Get a Network Interface resource from AWS. Supply one of the inputs as the filter.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS Network Interface id to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_network_interfaces

    """
    result = dict(comment=[], ret=None, result=True)

    if resource_id:
        ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
            ctx,
            NetworkInterfaceIds=[resource_id],
            Filters=filters,
        )
    else:
        ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
            ctx,
            Filters=filters,
        )

    # If there was an error in the call then report failure
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    present_states = hub.tool.aws.ec2.network_interface.convert_to_present(ctx, ret.ret)

    # If the resource can't be found but there were no results then "result" is True and "ret" is None
    if not present_states:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.network_interface", name=name
            )
        )
        return result

    # return the first result as a plain dictionary
    result["ret"] = next(iter((present_states).values()))

    return result


async def list_(hub, ctx, name: str = None, filters: List = None) -> Dict:
    """
    Use an un-managed Network INterface as a data-source. Supply one of the inputs as the filter.

    Args:
        name(string): The name of the Idem state.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_network_interfaces

    """

    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
        ctx,
        Filters=filters,
    )

    # If there was an error in the call then report failure
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    present_states = hub.tool.aws.ec2.network_interface.convert_to_present(ctx, ret.ret)
    if not present_states:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.network_interface", name=name
            )
        )
        return result

    # Return a list of dictionaries with details about all the instances
    result["ret"] = list(present_states.values())

    return result



from typing import Dict
from typing import List


async def search_raw(
    hub,
    ctx,
    zone_names: List = None,
    zone_ids: List = None,
    all_availability_zones: bool = None,
    filters: List = None,
) -> Dict:
    """
    Fetch one or more availability zones from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        hub: The redistributed pop central hub.
        ctx: The context.
        zone_names(List, optional): The names of the AWS Availability Zones, Local Zones, and Wavelength Zones to search.
        zone_ids(List, optional): The ids of the AWS Availability Zones, Local Zones, and Wavelength Zones to search.
        all_availability_zones(bool, optional): Indicate whether to include all AWS Availability Zones, Local Zones, and Wavelength Zones regardless of the opt-in status.
        filters(list, optional): One or more filters. For example: {"Name": "state", "Values": ["available"]},
          A complete list of filters can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_availability_zones

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}
    """
    result = dict(comment=[], ret=None, result=True)

    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )

    ret = await hub.exec.boto3.client.ec2.describe_availability_zones(
        ctx,
        Filters=boto3_filter,
        ZoneNames=zone_names,
        ZoneIds=zone_ids,
        AllAvailabilityZones=all_availability_zones,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def update(
    hub,
    ctx,
    name: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
    scope: str,
    resource_id: str,
    lock_token: str,
):
    """

    Args:
        hub:
        ctx:
        name: Name of resource going to update.
        raw_resource: Old state of resource or existing resource details.
        resource_parameters: Parameters from sls file
        scope: Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
        resource_id: The unique identifier for the web ACL.
        lock_token: A token used for optimistic locking.

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": {}}
    """

    parameters = OrderedDict(
        {
            "Name": "name",
            "DefaultAction": "default_action",
            "VisibilityConfig": "visibility_config",
            "Description": "description",
            "Rules": "rules",
            "CustomResponseBodies": "custom_response_bodies",
            "CaptchaConfig": "captcha_config",
        }
    )
    parameters_to_update = {}
    result = dict(comment=(), result=True, ret=None)
    resource_parameters.pop("Tags", None)

    for key, value in resource_parameters.items():
        if value is None or value == raw_resource[key]:
            continue
        parameters_to_update[key] = resource_parameters[key]

    if parameters_to_update:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.wafv2.update_web_acl(
                ctx,
                Scope=scope,
                Id=resource_id,
                LockToken=lock_token,
                **resource_parameters,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
            result["comment"] = result["comment"] + (f"Updated '{name}'",)

        result["ret"] = {}
        for parameter_raw, parameter_present in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

    return result



from typing import Dict


async def get(
    hub,
    ctx,
    resource_id: str,
) -> Dict:
    """
    Pass required params to get a ElasticLoadBalancing Load Balancer resource.

    Args:
        resource_id(Text): The name of the ElasticLoadBalancing Load Balancer. This name must be unique
            within your set of load balancers for the region, must have a maximum of 32 characters,

    Returns:
        {"result": True|False, "comment": A message List, "ret": None}

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.elb.load_balancer.search_raw(
        ctx=ctx, resource_id=resource_id
    )
    if not ret["result"]:
        if "LoadBalancerNotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.elb.load_balancer", name=resource_id
                )
            )
            result["comment"] += ret["comment"]
            return result
        result["comment"] += ret["comment"]
        result["result"] = False
        return result
    if not ret["ret"]["LoadBalancerDescriptions"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.elb.load_balancer", name=resource_id
            )
        )
        return result

    tags = attributes = []
    tags_ret = await hub.exec.boto3.client.elb.describe_tags(
        ctx, LoadBalancerNames=[resource_id]
    )
    if not tags_ret["result"]:
        result["comment"] = list(tags_ret["comment"])
        result["result"] = False
        return result
    else:
        if tags_ret.get("ret") and tags_ret.get("ret")["TagDescriptions"]:
            tags = (tags_ret["ret"]["TagDescriptions"][0]).get("Tags")

    attributes_ret = await hub.exec.boto3.client.elb.describe_load_balancer_attributes(
        ctx, LoadBalancerName=resource_id
    )
    if not attributes_ret["result"]:
        result["comment"] = list(attributes_ret["comment"])
        result["result"] = False
        return result
    else:
        if attributes_ret.get("ret") and attributes_ret.get("ret").get(
            "LoadBalancerAttributes"
        ):
            attributes = attributes_ret["ret"].get("LoadBalancerAttributes")

    result[
        "ret"
    ] = hub.tool.aws.elb.conversion_utils.convert_raw_load_balancer_to_present(
        raw_resource=ret["ret"]["LoadBalancerDescriptions"][0],
        idem_resource_name=resource_id,
        tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
        attributes=attributes,
    )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def convert_raw_key_to_present(
    hub, ctx, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    describe_parameters = OrderedDict(
        {
            "KeyId": "resource_id",
            "KeyArn": "arn",
            "Arn": "arn",
            "KeyState": "key_state",
            "Description": "description",
            "KeyUsage": "key_usage",
            "KeySpec": "key_spec",
            "MultiRegion": "multi_region",
            "Policy": "policy",
            "BypassPolicyLockoutSafetyCheck": "bypass_policy_lockout_safety_check",
        }
    )

    translated_resource = {}
    for parameter_raw, parameter_present in describe_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    # Get key tags
    # For tag operations key_id is used and not resource_i
    key_tags = await hub.exec.boto3.client.kms.list_resource_tags(
        ctx, KeyId=translated_resource["resource_id"]
    )
    if key_tags and key_tags["result"] is True:
        translated_resource[
            "tags"
        ] = hub.tool.aws.tag_utils.convert_tag_list_to_dict_tagkey(
            key_tags["ret"].get("Tags", [])
        )
    else:
        raise Exception(key_tags["comment"])

    # Get key policy
    key_policy = await hub.exec.boto3.client.kms.get_key_policy(
        ctx, KeyId=translated_resource["resource_id"], PolicyName="default"
    )
    if (
        key_policy
        and key_policy["result"] is True
        and key_policy["ret"].get("Policy", None)
    ):
        translated_resource[
            "policy"
        ] = hub.tool.aws.state_comparison_utils.standardise_json(
            key_policy["ret"].get("Policy")
        )
    else:
        raise Exception(key_policy["comment"])

    # Get key rotation value if key is Enabled (not if Disabled or PendingDeletion)
    if translated_resource["key_state"] == "Enabled":
        key_rotation = await hub.exec.boto3.client.kms.get_key_rotation_status(
            ctx, KeyId=translated_resource["resource_id"]
        )
        if key_rotation["result"] is True:
            translated_resource["enable_key_rotation"] = key_rotation["ret"].get(
                "KeyRotationEnabled", False
            )
        else:
            raise Exception(key_rotation["comment"])

    return translated_resource


def convert_raw_key_alias_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:

    # Arn is not used for present but required for arg binding
    resource_parameters = OrderedDict(
        {
            "AliasArn": "arn",
            "TargetKeyId": "target_key_id",
        }
    )
    # AliasName is the unique identifier for KMS Alias, so it is set as resource_id
    translated_resource = {"resource_id": raw_resource.get("AliasName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    return translated_resource



from typing import Dict
from typing import List


async def get(
    hub,
    ctx,
    name,
    resource_id: str = None,
    filters: List[Dict[str, str]] = None,
) -> Dict:
    """
    Supply one of the inputs as the filter to find a key pair.
    If more than one resource is found, the first resource returned from AWS will be used.
    The function returns None when no resource is found.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS key pair id to identify the resource.
        filters(list, optional): One or more filters. For example: {"Name": "key-pair-id", "Values": [resource_id]},
        A complete list of filters can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_key_pairs
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.ec2.key_pair.search_raw(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        filters=filters,
    )
    if not ret["result"]:
        if "InvalidKeyPair.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.key_pair", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["KeyPairs"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.key_pair", name=name
            )
        )
        return result

    if len(ret["ret"]["KeyPairs"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type="aws.ec2.key_pair", name=name
            )
        )
    resource = ret["ret"]["KeyPairs"][0]
    result["ret"] = hub.tool.aws.ec2.conversion_utils.convert_raw_key_pair_to_present(
        raw_resource=resource
    )
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions for AWS DynamoDB Table resources.
"""


def convert_raw_table_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    r"""
    Util functions to convert raw Dynamo DB resource state from AWS to present input format.

    Args:
        hub: required for functions in hub.
        ctx: context.
        raw_resource(Dict[str, Any]): The AWS response to convert.
        idem_resource_name(string): The name of the Idem state.

    Returns:
        Dict[str, Any]: Common idem present state
    """
    describe_parameters = OrderedDict(
        {
            "AttributeDefinitions": "attribute_definitions",
            "KeySchema": "key_schema",
            "TableArn": "table_arn",
            "TableStatus": "table_status",
        }
    )
    new_table_resource = {"resource_id": idem_resource_name, "name": idem_resource_name}
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if raw_resource.get(parameter_old_key) is not None:
            new_table_resource[parameter_new_key] = raw_resource.get(parameter_old_key)

    if raw_resource.get("BillingModeSummary"):
        new_table_resource["billing_mode"] = raw_resource.get("BillingModeSummary").get(
            "BillingMode"
        )
    else:
        new_table_resource["billing_mode"] = "PROVISIONED"

    if raw_resource.get("GlobalSecondaryIndexes"):
        global_secondary_indexes = []
        for global_index in raw_resource.get("GlobalSecondaryIndexes"):
            new_global_index = {
                "IndexName": global_index["IndexName"],
                "KeySchema": global_index["KeySchema"],
                "Projection": global_index["Projection"],
            }
            if global_index.get("ProvisionedThroughput"):
                provisioned_throughput = global_index.get("ProvisionedThroughput")
                new_provisioned = {
                    "ReadCapacityUnits": provisioned_throughput["ReadCapacityUnits"],
                    "WriteCapacityUnits": provisioned_throughput["WriteCapacityUnits"],
                }
                new_global_index["ProvisionedThroughput"] = new_provisioned
            global_secondary_indexes.append(new_global_index)
        new_table_resource["global_secondary_indexes"] = global_secondary_indexes

    if raw_resource.get("LocalSecondaryIndexes"):
        local_secondary_indexes = []
        for local_index in raw_resource.get("LocalSecondaryIndexes"):
            new_local_index = {
                "IndexName": local_index["IndexName"],
                "KeySchema": local_index["KeySchema"],
                "Projection": local_index["Projection"],
            }
            local_secondary_indexes.append(new_local_index)
        new_table_resource["local_secondary_indexes"] = local_secondary_indexes

    continuous_backups_description = raw_resource.get("ContinuousBackupsDescription")
    if continuous_backups_description:
        point_in_time_recovery_description = continuous_backups_description.get(
            "PointInTimeRecoveryDescription"
        )
        if point_in_time_recovery_description:
            if (
                point_in_time_recovery_description.get(
                    "PointInTimeRecoveryStatus", "DISABLED"
                )
                == "ENABLED"
            ):
                point_in_time_recovery = {"Enabled": True}
            else:
                point_in_time_recovery = {"Enabled": False}
            new_table_resource["point_in_time_recovery"] = point_in_time_recovery

    if raw_resource.get("ProvisionedThroughput"):
        provisioned_throughput = raw_resource.get("ProvisionedThroughput")
        new_provisioned = {
            "ReadCapacityUnits": provisioned_throughput["ReadCapacityUnits"],
            "WriteCapacityUnits": provisioned_throughput["WriteCapacityUnits"],
        }
        new_table_resource["provisioned_throughput"] = new_provisioned

    if raw_resource.get("Replicas"):
        new_table_resource["replica_updates"] = raw_resource.get("Replicas")

    if raw_resource.get("SSEDescription"):
        sse_specification = raw_resource.get("SSEDescription")
        new_sse_specification = {}
        kms_master_key_arn = sse_specification.get("KMSMasterKeyArn")
        if kms_master_key_arn:
            new_sse_specification["KMSMasterKeyId"] = kms_master_key_arn
        sse_type = sse_specification.get("SSEType")
        if sse_type:
            new_sse_specification["SSEType"] = sse_type
        if (
            sse_specification.get("Status") == "ENABLED"
            or sse_specification.get("Status") == "ENABLING"
        ):
            new_sse_specification["Enabled"] = True
        else:
            new_sse_specification["Enabled"] = False
        new_table_resource["sse_specification"] = new_sse_specification
    else:
        new_table_resource["sse_specification"] = {
            "Enabled": False,
        }

    if raw_resource.get("StreamSpecification"):
        new_table_resource["stream_specification"] = raw_resource.get(
            "StreamSpecification"
        )

    if raw_resource.get("TableClassSummary"):
        new_table_resource["table_class"] = raw_resource.get("TableClassSummary").get(
            "TableClass"
        )

    tags = raw_resource.get("Tags")
    if tags:
        new_table_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            tags
        )

    time_to_live_description = raw_resource.get("TimeToLiveDescription")
    if time_to_live_description:
        if time_to_live_description.get("TimeToLiveStatus", "DISABLED") == "ENABLED":
            time_to_live = {
                "Enabled": True,
                "AttributeName": time_to_live_description.get("AttributeName"),
            }
        else:
            time_to_live = {"Enabled": False}
        new_table_resource["time_to_live"] = time_to_live

    return new_table_resource


async def update(
    hub,
    ctx,
    name: str,
    resource_id: str,
    old_state: Dict[str, Any],
    desired_state: Dict[str, Any],
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    Updates a DynamoDB Table resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        name(string): The name of the Idem state.
        resource_id(string): The name of the Dynamodb table.
        old_state(dict): The old state for the specified table.
        desired_state(dict): The desired state for the specified table.
        timeout(Dict, optional): Timeout configuration for updates.

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)
    updated_attributes = {}

    table_parameters = OrderedDict(
        {
            "attribute_definitions": "AttributeDefinitions",
            "billing_mode": "BillingMode",
            "global_secondary_indexes": "GlobalSecondaryIndexUpdates",
            "provisioned_throughput": "ProvisionedThroughput",
            "replica_updates": "ReplicaUpdates",
            "stream_specification": "StreamSpecification",
            "sse_specification": "SSESpecification",
            "table_class": "TableClass",
        }
    )
    table_parameters_to_update = {}

    attribute_definitions = desired_state.get("attribute_definitions")
    if attribute_definitions is not None:
        if not hub.tool.aws.state_comparison_utils.are_lists_identical(
            old_state.get("attribute_definitions"),
            attribute_definitions,
        ):
            table_parameters_to_update["AttributeDefinitions"] = attribute_definitions

    global_secondary_indexes = desired_state.get("global_secondary_indexes")
    if global_secondary_indexes is not None:
        if not hub.tool.aws.state_comparison_utils.are_lists_identical(
            old_state.get("global_secondary_indexes"),
            global_secondary_indexes,
        ):
            table_parameters_to_update[
                "GlobalSecondaryIndexUpdates"
            ] = global_secondary_indexes

    replica_updates = desired_state.get("replica_updates")
    if replica_updates is not None:
        if not hub.tool.aws.state_comparison_utils.are_lists_identical(
            old_state.get("replica_updates"),
            replica_updates,
        ):
            table_parameters_to_update["ReplicaUpdates"] = replica_updates

    for key, value in desired_state.items():
        if key in table_parameters.keys():
            if value is not None and value != old_state.get(key):
                table_parameters_to_update[table_parameters[key]] = value

    if table_parameters_to_update:
        for parameter_present, parameter_raw in table_parameters.items():
            if parameter_raw in table_parameters_to_update:
                updated_attributes[parameter_present] = table_parameters_to_update[
                    parameter_raw
                ]

        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.dynamodb.update_table(
                ctx=ctx,
                TableName=resource_id,
                **table_parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result

            hub.log.debug(f"Waiting on updating aws.dynamodb.table '{name}'")
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("update") if timeout else None,
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "dynamodb",
                    "table_exists",
                    TableName=resource_id,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["result"] = False
                result["comment"] += (str(e),)

    if desired_state.get("point_in_time_recovery"):
        update_ret = await hub.tool.aws.dynamodb.table.update_point_in_time_recovery(
            ctx,
            name=name,
            resource_id=resource_id,
            old_point_in_time_recovery=old_state.get("point_in_time_recovery"),
            new_point_in_time_recovery=desired_state.get("point_in_time_recovery"),
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + update_ret["comment"]
            return result

        if update_ret["ret"]:
            updated_attributes["point_in_time_recovery"] = copy.deepcopy(
                update_ret["ret"]
            )

    if desired_state.get("time_to_live"):
        update_ret = await hub.tool.aws.dynamodb.table.update_time_to_live(
            ctx,
            name=name,
            resource_id=resource_id,
            old_time_to_live=old_state.get("time_to_live"),
            new_time_to_live=desired_state.get("time_to_live"),
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + update_ret["comment"]
            return result

        if update_ret["ret"]:
            updated_attributes["time_to_live"] = copy.deepcopy(update_ret["ret"])

    if updated_attributes:
        result["ret"] = copy.deepcopy(updated_attributes)
        if ctx.get("test", False):
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.would_update_comment(
                "aws.dynamodb.table", name
            )
        else:
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.update_comment("aws.dynamodb.table", name)

    return result


async def update_point_in_time_recovery(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    old_point_in_time_recovery: Dict = None,
    new_point_in_time_recovery: Dict = None,
) -> Dict[str, Any]:
    r"""
    Updates the point in time recovery settings for a DynamoDB Table resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        name(string): The name of the Idem state.
        resource_id(string, optional): The name of the Dynamodb table.
        old_point_in_time_recovery(dict, optional): The old point in time recovery settings for the specified table.
        new_point_in_time_recovery(dict, optional): The new point in time recovery settings for the specified table.

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)

    if (
        new_point_in_time_recovery
        and new_point_in_time_recovery != old_point_in_time_recovery
    ):
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.dynamodb.update_continuous_backups(
                ctx,
                TableName=resource_id,
                PointInTimeRecoverySpecification={
                    "PointInTimeRecoveryEnabled": new_point_in_time_recovery.get(
                        "Enabled", False
                    ),
                },
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result
        result["ret"] = new_point_in_time_recovery

    return result


async def update_time_to_live(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    old_time_to_live: Dict = None,
    new_time_to_live: Dict = None,
) -> Dict[str, Any]:
    r"""
    Updates the time to live (TTL) settings for a DynamoDB Table resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        name(string): The name of the Idem state.
        resource_id(string, optional): The name of the Dynamodb table.
        old_time_to_live(dict, optional): The old time to live (TTL) settings for the specified table.
        new_time_to_live(dict, optional): The new time to live (TTL) settings for the specified table.

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)

    if new_time_to_live and (
        (not old_time_to_live and new_time_to_live.get("Enabled", False))
        or (old_time_to_live and new_time_to_live != old_time_to_live)
    ):
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.dynamodb.update_time_to_live(
                ctx,
                TableName=resource_id,
                TimeToLiveSpecification=new_time_to_live,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result
        result["ret"] = new_time_to_live

    return result



from typing import Any
from typing import Dict


async def move_account(
    hub, ctx, account_id, result: Dict[str, Any], destination_parent_id
) -> Dict[str, Any]:
    if destination_parent_id is not None:

        parents = await hub.exec.boto3.client.organizations.list_parents(
            ctx, ChildId=account_id
        )
        result["result"] = result["result"] and parents["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + parents["comment"]
            return result

        if parents:
            parent_id = parents["ret"]["Parents"][0]["Id"]

            if parent_id == destination_parent_id:
                result["comment"] = result["comment"] + (
                    f"aws.organizations.account {account_id} already at {destination_parent_id}",
                )
                return result
            else:

                account_move_ret = (
                    await hub.exec.boto3.client.organizations.move_account(
                        ctx,
                        AccountId=account_id,
                        SourceParentId=parent_id,
                        DestinationParentId=destination_parent_id,
                    )
                )
                if not account_move_ret:
                    result["comment"] = result["comment"] + (
                        f"Could not move aws.organizations.account {account_id} to {destination_parent_id} Error: {account_move_ret['comment']}",
                    )

                    result["result"] = False

                    return result
                result["comment"] = result["comment"] + (
                    f"Successfully moved aws.organizations.account {account_id} under {destination_parent_id}",
                )
                return result



from typing import Any
from typing import Dict
from typing import List

import dict_tools


async def update(
    hub,
    ctx,
    name: str,
    resource_id: str,
    before: Dict[str, Any],
    auto_scaling_group_name: str,
    policy_type: str,
    adjustment_type: str,
    min_adjustment_step: int,
    min_adjustment_magnitude: int,
    scaling_adjustment: int,
    cooldown: int,
    metric_aggregation_type: str,
    step_adjustments: List[Dict[str, Any]],
    estimated_instance_warmup: int,
    target_tracking_configuration: Dict[str, Any],
    enabled: bool,
    predictive_scaling_configuration: Dict[str, Any],
):
    result = dict(comment=(), result=True, ret=None)
    update_allowed_params = [
        "policy_type",
        "adjustment_type",
        "min_adjustment_step",
        "min_adjustment_magnitude",
        "scaling_adjustment",
        "cooldown",
        "metric_aggregation_type",
        "estimated_instance_warmup",
        "enabled",
    ]

    modified_params = {}
    for parameter in update_allowed_params:
        input_parameter_value = locals()[parameter]
        if (
            input_parameter_value is not None
            and before.get(parameter) != input_parameter_value
        ):
            modified_params[parameter] = input_parameter_value
    if dict_tools.data.recursive_diff(
        step_adjustments,
        before.get("step_adjustments"),
        ignore_order=True,
    ):
        modified_params["step_adjustments"] = step_adjustments
    if dict_tools.data.recursive_diff(
        before.get("target_tracking_configuration"),
        target_tracking_configuration,
        ignore_order=True,
    ):
        modified_params["target_tracking_configuration"] = target_tracking_configuration
    if dict_tools.data.recursive_diff(
        before.get("predictive_scaling_configuration"),
        predictive_scaling_configuration,
        ignore_order=True,
    ):
        modified_params[
            "predictive_scaling_configuration"
        ] = predictive_scaling_configuration

    if modified_params:
        result["ret"] = modified_params
        if not ctx.get("test", False):
            ret = await hub.exec.boto3.client.autoscaling.put_scaling_policy(
                ctx,
                AutoScalingGroupName=auto_scaling_group_name,
                PolicyName=resource_id,
                PolicyType=policy_type,
                AdjustmentType=adjustment_type,
                MinAdjustmentStep=min_adjustment_step,
                MinAdjustmentMagnitude=min_adjustment_magnitude,
                ScalingAdjustment=scaling_adjustment,
                Cooldown=cooldown,
                MetricAggregationType=metric_aggregation_type,
                StepAdjustments=step_adjustments,
                EstimatedInstanceWarmup=estimated_instance_warmup,
                TargetTrackingConfiguration=target_tracking_configuration,
                Enabled=enabled,
                PredictiveScalingConfiguration=predictive_scaling_configuration,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.autoscaling.scaling_policy", name=name
            )
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.elbv2.add_listener_certificates
hub.exec.boto3.client.elbv2.create_listener
hub.exec.boto3.client.elbv2.delete_listener
hub.exec.boto3.client.elbv2.describe_listener_certificates
hub.exec.boto3.client.elbv2.describe_listeners
hub.exec.boto3.client.elbv2.describe_rules
hub.exec.boto3.client.elbv2.modify_listener
hub.exec.boto3.client.elbv2.remove_listener_certificates
hub.exec.boto3.client.elbv2.remove_tags
hub.exec.boto3.client.elbv2.describe_tags
hub.exec.boto3.client.elbv2.add_tags


"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    lb_arn: str,
    default_actions: List[
        make_dataclass(
            "Action",
            [
                ("Type", str),
                ("TargetGroupArn", str, field(default=None)),
                (
                    "AuthenticateOidcConfig",
                    make_dataclass(
                        "AuthenticateOidcActionConfig",
                        [
                            ("Issuer", str),
                            ("AuthorizationEndpoint", str),
                            ("TokenEndpoint", str),
                            ("UserInfoEndpoint", str),
                            ("ClientId", str),
                            ("ClientSecret", str, field(default=None)),
                            ("SessionCookieName", str, field(default=None)),
                            ("Scope", str, field(default=None)),
                            ("SessionTimeout", int, field(default=None)),
                            (
                                "AuthenticationRequestExtraParams",
                                Dict[str, str],
                                field(default=None),
                            ),
                            ("OnUnauthenticatedRequest", str, field(default=None)),
                            ("UseExistingClientSecret", bool, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "AuthenticateCognitoConfig",
                    make_dataclass(
                        "AuthenticateCognitoActionConfig",
                        [
                            ("UserPoolArn", str),
                            ("UserPoolClientId", str),
                            ("UserPoolDomain", str),
                            ("SessionCookieName", str, field(default=None)),
                            ("Scope", str, field(default=None)),
                            ("SessionTimeout", int, field(default=None)),
                            (
                                "AuthenticationRequestExtraParams",
                                Dict[str, str],
                                field(default=None),
                            ),
                            ("OnUnauthenticatedRequest", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("Order", int, field(default=None)),
                (
                    "RedirectConfig",
                    make_dataclass(
                        "RedirectActionConfig",
                        [
                            ("StatusCode", str),
                            ("Protocol", str, field(default=None)),
                            ("Port", str, field(default=None)),
                            ("Host", str, field(default=None)),
                            ("Path", str, field(default=None)),
                            ("Query", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "FixedResponseConfig",
                    make_dataclass(
                        "FixedResponseActionConfig",
                        [
                            ("StatusCode", str),
                            ("MessageBody", str, field(default=None)),
                            ("ContentType", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "ForwardConfig",
                    make_dataclass(
                        "ForwardActionConfig",
                        [
                            (
                                "TargetGroups",
                                List[
                                    make_dataclass(
                                        "TargetGroupTuple",
                                        [
                                            (
                                                "TargetGroupArn",
                                                str,
                                                field(default=None),
                                            ),
                                            ("Weight", int, field(default=None)),
                                        ],
                                    )
                                ],
                                field(default=None),
                            ),
                            (
                                "TargetGroupStickinessConfig",
                                make_dataclass(
                                    "TargetGroupStickinessConfig",
                                    [
                                        ("Enabled", bool, field(default=None)),
                                        ("DurationSeconds", int, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
            ],
        )
    ],
    resource_id: str = None,
    protocol: str = None,
    port: int = None,
    ssl_policy: str = None,
    default_certificates: List[
        make_dataclass(
            "Certificate",
            [
                ("CertificateArn", str, field(default=None)),
            ],
        )
    ] = None,
    certificates: List[
        make_dataclass(
            "Certificate",
            [
                ("CertificateArn", str, field(default=None)),
            ],
        )
    ] = None,
    alpn_policy: List[str] = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    """Creates a listener for the specified Application Load Balancer, Network Load Balancer, or Gateway Load Balancer.
    For more information, see the following:    Listeners for your Application Load Balancers     Listeners for your
    Network Load Balancers     Listeners for your Gateway Load Balancers    This operation is idempotent, which
    means that it completes at most one time. If you attempt to create multiple listeners with the same settings,
    each call succeeds.

    Args:
        name(str): IDEM name of the state.

        resource_id(str, optional):
            An identifier of the resource in the provider. Defaults to None. The Amazon Resource Name (ARN) of the
            ElasticLoadBalancingv2 listener.

        lb_arn(str):
            The Amazon Resource Name (ARN) of the load balancer.

        protocol(str, optional):
            The protocol for connections from clients to the load balancer. For Application Load Balancers,
            the supported protocols are HTTP and HTTPS. For Network Load Balancers, the supported protocols
            are TCP, TLS, UDP, and TCP_UDP. You can’t specify the UDP or TCP_UDP protocol if dual-stack mode
            is enabled. You cannot specify a protocol for a Gateway Load Balancer. Defaults to None.

        port(int, optional):
            The port on which the load balancer is listening. You cannot specify a port for a Gateway LoadBalancer.
            Defaults to None.

        ssl_policy(str, optional):
            [HTTPS and TLS listeners] The security policy that defines which protocols and ciphers are supported.
            For more information, see Security policies in the Application Load Balancers Guide and Security policies
            in the Network Load Balancers Guide. Defaults to None.

        default_certificates(list[dict[str, Any]], optional):
            [HTTPS and TLS listeners] The default certificate for the listener. You must provide exactly one
            certificate. Set CertificateArn to the certificate ARN but do not set IsDefault. Defaults to None.
                * CertificateArn (str, optional):
                    The Amazon Resource Name (ARN) of the certificate.

        certificates(list[dict[str, Any]], optional):
            The certificate to add. You can specify one certificate per call. Set CertificateArn to the certificate ARN
            but do not set IsDefault .
                * CertificateArn (str, optional): The Amazon Resource Name (ARN) of the certificate.

        default_actions(list[dict[str, Any]]): The actions for the default rule.
            * Type (str):
                The type of action.

            * TargetGroupArn (str, optional):
                The Amazon Resource Name (ARN) of the target group. Specify only when Type is forward and you want to
                route to a single target group. To route to one or more target groups, use ForwardConfig instead.

            * AuthenticateOidcConfig (dict[str, Any], optional):
                [HTTPS listeners] Information about an identity provider that is compliant with OpenID Connect (OIDC).
                Specify only when Type is authenticate-oidc.

                    * Issuer (str):
                        The OIDC issuer identifier of the IdP. This must be a full URL, including the HTTPS protocol,
                        the domain, and the path.

                    * AuthorizationEndpoint (str):
                        The authorization endpoint of the IdP. This must be a full URL, including the HTTPS protocol,
                        the domain, and the path.

                    * TokenEndpoint (str):
                        The token endpoint of the IdP. This must be a full URL, including the HTTPS protocol, the
                        domain, and the path.

                    * UserInfoEndpoint (str):
                        The user info endpoint of the IdP. This must be a full URL, including the HTTPS protocol, the
                        domain, and the path.

                    * ClientId (str):
                        The OAuth 2.0 client identifier.

                    * ClientSecret (str, optional):
                        The OAuth 2.0 client secret. This parameter is required if you are creating a rule. If you are
                        modifying a rule, you can omit this parameter if you set UseExistingClientSecret to true.

                    * SessionCookieName (str, optional):
                        Name of the cookie used to maintain session information. The default is AWSELBAuthSessionCookie.

                    * Scope (str, optional):
                        The set of user claims to be requested from the IdP. The default is openid. To verify which
                        scope values your IdP supports and how to separate multiple values, see the documentation for
                        your IdP.

                    * SessionTimeout (int, optional):
                        Maximum duration of authentication session, in seconds. The default is 604800 seconds (7days).

                    * AuthenticationRequestExtraParams (Dict[str, str], optional):
                        The query parameters (up to 10) to include in the redirect request to the authorization endpoint.

                    * OnUnauthenticatedRequest (str, optional):
                        The behavior if the user is not authenticated. The following are possible values:
                            deny - Return an HTTP 401 Unauthorized error.
                            allow - Allow the request to be forwarded to the target.
                            authenticate - Redirect the request to IdP authorization endpoint. This is the default value.

                    * UseExistingClientSecret (bool, optional):
                        Indicates whether to use the existing client secret when modifying a rule. If you are creating
                        a rule, you can omit this parameter or set it to false.

            * AuthenticateCognitoConfig (dict[str, Any], optional):
                [HTTPS listeners] Information for using Amazon Cognito to authenticate users. Specify only when Type is
                authenticate-cognito.

                    * UserPoolArn (str):
                        The Amazon Resource Name (ARN) of the Amazon Cognito user pool.

                    * UserPoolClientId (str):
                        The ID of the Amazon Cognito user pool client.

                    * UserPoolDomain (str):
                        The domain prefix or fully-qualified domain name of Amazon Cognito user pool.

                    * SessionCookieName (str, optional):
                        Name of the cookie used to maintain session information. Defaults to AWSELBAuthSessionCookie.

                    * Scope (str, optional):
                        The set of user claims to be requested from the IdP. The default is openid. To verify which
                        scope values your IdP supports and how to separate multiple values, see the documentation for
                        your IdP.

                    * SessionTimeout (int, optional):
                        Maximum duration of the authentication session, in seconds. Defaults to 604800 seconds (7days).

                    * AuthenticationRequesstrraParams (dict[str, str], optional):
                        The query parameters (up to 10) to include in redirect request to authorization endpoint.

                    * OnUnauthenticatedRequest (str, optional):
                        The behavior if the user is not authenticated. The following are possible values:
                            deny - Return an HTTP 401 Unauthorized error.
                            allow - Allow the request to be forwarded to the target.
                            authenticate - Redirect the request to the IdP authorization endpoint. This is the default value.

            * Order (int, optional):
                The order for the action. This value is required for rules with multiple actions. The action with the
                lowest value for order is performed first.

            * RedirectConfig (dict[str, Any], optional):
                [Application Load Balancer] Information for creating a redirect action. Specify only when Type is redirect.

                * Protocol (str, optional):
                    The protocol. You can specify HTTP, HTTPS, or #{protocol}. You can redirect HTTP to HTTP, HTTP to
                    HTTPS, and HTTPS to HTTPS. You cannot redirect HTTPS to HTTP.

                * Port (str, optional):
                    The port. You can specify a value from 1 to 65535 or #{port}.

                * Host (str, optional):
                    The hostname. This component is not percent-encoded. The hostname can contain #{host}.

                * Path (str, optional):
                    The absolute path, starting with the leading "/". This component is not percent-encoded. The path
                    can contain #{host}, #{path}, and #{port}.

                * Query (str, optional):
                    The query parameters, URL-encoded when necessary, but not percent-encoded. Do not include the
                    leading "?", as it is automatically added. You can specify any of the reserved keywords.

                * StatusCode (str):
                    HTTP redirect code. Redirect is either permanent (HTTP 301) or temporary (HTTP 302).

            * FixedResponseConfig (dict[str, Any], optional):
                [Application Load Balancer] Information for creating an action that returns a custom HTTP response.
                Specify only when Type is fixed-response.

                    * MessageBody (str, optional): The message.

                    * StatusCode (str): The HTTP response code (2XX, 4XX, or 5XX).

                    * ContentType (str, optional):
                        The content type. Valid Values: str/plain | str/css | str/html | application/javascript
                        | application/json

            * ForwardConfig (dict[str, Any], optional):
                Information for creating an action that distributes requests among one or more target groups. For
                Network Load Balancers, you can specify a single target group. Specify only when Type is forward. If you
                specify both ForwardConfig and TargetGroupArn, you can specify only one target group using ForwardConfig
                and it must be the same target group specified in TargetGroupArn.

                * TargetGroups (list[dict[str, Any]], optional):
                    One or more target groups. For Network Load Balancers, you can specify a single target group.
                        * TargetGroupArn (str, optional): The Amazon Resource Name (ARN) of the target group.
                        * Weight (int, optional): The weight. The range is 0 to 999.

                * TargetGroupStickinessConfig (dict[str, Any], optional):
                    The target group stickiness for the rule.
                        * Enabled (bool, optional): Indicates whether target group stickiness is enabled.
                        * DurationSeconds (int, optional): The time period, in seconds, during which requests from
                          a client should be routed to the same target group. The range is 1-604800 seconds (7 days).

        alpn_policy(list[str], optional):
            [TLS listeners] The name of the Application-Layer Protocol Negotiation (ALPN) policy. You can specify one
            policy name. The following are the possible values:
            HTTP1Only
            HTTP2Only
            HTTP2Optional
            HTTP2Preferred
            None
        tags(dict[str, str], optional):
            The tags to assign to the listener. Defaults to None.
                * Key (str): The key of the tag.
                * Value (str, optional): The value of the tag.

    Request Syntax:
        [listener-name]:
            aws.elbv2.listener.present:
                - name: "string"
                - lb_arn: "string"
                - default_actions: "list"
                - resource_id: "string"
                - protocol: "string"
                - port: "integer"
                - ssl_policy: "string"
                - default_certificates: "list"
                - certificates: "list"
                - alpn_policy: "list"
                - tags:
                    - 'string': "string"
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.elbv2.listener.present:
                - name: 'my-listener'
                - lb_arn: 'arn:aws:elasticloadbalancing:us-west-2:45678908712456908:loadbalancer/app/my-load-balancer/50ghtc66t95c0c9188'
                - default_actions:
                    - TargetGroupArn: arn:aws:elasticloadbalancing:us-west-2:45678908712456908:targetgroup/my-targets/73e2d6bc24d8a067
                      Type: Type
                - resource_id: 'arn:aws:elasticloadbalancing:us-west-2:45678908712456908:listener/app/my-load-balancer/50dc6c495c0c9188/f2f7dc8efc522ab2'
                - protocol: 'HTTPS'
                - port: 443
                - ssl_policy: 'ELBSecurityPolicy-2015-05'
                - default_certificates:
                    - CertificateArn: arn:aws:iam::45678908712456908:server-certificate/default-cert
                - certificates:
                    - CertificateArn: arn:aws:iam::45678908712456908:server-certificate/my-server-cert
                - alpn_policy:
                    - HTTP2Only
                    - HTTP2Preferred
                - tags:
                    name: my-listener

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    plan_state = {}
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.elbv2.listener.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = list(before["comment"])
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        input_params = {
            "port": port,
            "protocol": protocol,
            "ssl_policy": ssl_policy,
            "certificates": certificates,
            "default_certificates": default_certificates,
            "default_actions": default_actions,
            "tags": tags,
            "alpn_policy": alpn_policy,
            "listener_arn": resource_id,
        }
        update_return = await hub.tool.aws.elbv2.listener.update(
            ctx=ctx,
            name=name,
            current_state=result["old_state"],
            input_map=input_params,
            plan_state=plan_state,
            resource_id=resource_id,
        )
        result["comment"] = update_return["comment"]
        if not update_return["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_return["ret"])
        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += list(
                    hub.tool.aws.comment_utils.would_update_comment(
                        resource_type="aws.elbv2.listener", name=name
                    )
                )
            else:
                result["comment"] += list(
                    hub.tool.aws.comment_utils.update_comment(
                        resource_type="aws.elbv2.listener", name=name
                    )
                )

        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_tags = await hub.tool.aws.elbv2.tag.update(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=result["old_state"]["tags"],
                new_tags=tags,
            )
            result["comment"] += list(update_tags["comment"])
            result["result"] = update_tags["result"]
            resource_updated = resource_updated or bool(update_tags["ret"])
            if ctx.get("test", False) and update_tags["ret"]:
                plan_state["tags"] = update_tags["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "lb_arn": lb_arn,
                    "default_actions": default_actions,
                    "default_certificates": default_certificates,
                    "protocol": protocol,
                    "port": port,
                    "ssl_policy": ssl_policy,
                    "certificates": certificates,
                    "alpn_policy": alpn_policy,
                    "tags": tags,
                },
            )
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.elbv2.listener", name=name
                )
            )
            return result
        ret = await hub.exec.boto3.client.elbv2.create_listener(
            ctx,
            LoadBalancerArn=lb_arn,
            Protocol=protocol,
            Port=port,
            SslPolicy=ssl_policy,
            Certificates=default_certificates,
            DefaultActions=default_actions,
            AlpnPolicy=alpn_policy,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            if tags
            else None,
        )
        if not ret["result"] and not ret["ret"]:
            result["comment"] = list(ret["comment"])
            result["result"] = False
            return result
        resource_id = (ret["ret"]["Listeners"][0]).get("ListenerArn")
        result["comment"] = list(
            hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.elbv2.listener", name=name
            )
        )
        if certificates:
            for certificate in certificates:
                add_certificate = (
                    await hub.exec.boto3.client.elbv2.add_listener_certificates(
                        ctx, ListenerArn=resource_id, Certificates=[certificate]
                    )
                )
                if not add_certificate["result"]:
                    result["comment"] += list(add_certificate["comment"])
                    result["result"] = False
                    return result
            result["comment"].append("Added Certificates.")

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.elbv2.listener.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] += after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    Deletes the specified listener. Alternatively, your listener is deleted when you delete the load balancer
    to which it is attached.

    Args:
        name(str):
            Idem name of the ElasticLoadBalancingv2 Listener.

        resource_id(str, optional):
            The Amazon Resource Name (ARN) of the ElasticLoadBalancingv2 listener.

    Request Syntax:
        [listener-name]:
          aws.elbv2.listener.absent:
            - name: "string"
            - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_listener_name:
                aws.elbv2.listener.absent:
                  - name: test_listener_name
                  - resource_id: arn:aws:elasticloadbalancing:ua-west-2:45678908712456908:listener/app/my-load-balancer/50dc6c495c0c9188/f2f7dc8efc522ab2
    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elbv2.listener", name=name
            )
        )
        return result
    before = await hub.exec.aws.elbv2.listener.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = list(before["comment"])
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elbv2.listener", name=name
            )
        )
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_delete_comment(
                    resource_type="aws.elbv2.listener", name=name
                )
            )
            return result
        else:
            ret = await hub.exec.boto3.client.elbv2.delete_listener(
                ctx, ListenerArn=resource_id
            )
            if not ret["result"]:
                result["comment"] = list(before["comment"])
                result["result"] = False
                return result
            result["comment"] = list(
                hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.elbv2.listener", name=name
                )
            )
    return result


async def describe(
    hub,
    ctx,
) -> Dict[str, Dict[str, Any]]:
    r"""
    1. Describe the resource in a way that can be recreated/managed with the corresponding "present" function.
    2. Describes the specified listeners or the listeners for the specified Application Load Balancer, Network Load
       Balancer, or Gateway Load Balancer. You must specify either a load balancer or one or more listeners.
    3. Users must specify one of the 2 params: list of listener ARNs or a load balancer ARN. Else, describe operation
       fails with the following error.
            'ClientError: An error occurred (ValidationError) when calling the DescribeListeners operation: You must
            specify either listener ARNs or a load balancer ARN'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

        $ idem describe aws.elbv2.listener
    """
    result = {}
    load_balancer_ret = await hub.exec.boto3.client.elbv2.describe_load_balancers(ctx)
    if not load_balancer_ret["result"]:
        hub.log.debug(
            f"Could not describe aws.elbv2.load_balancer: {load_balancer_ret['comment']}"
        )
        return result
    if load_balancer_ret["ret"].get("LoadBalancers"):
        for load_balancer in load_balancer_ret["ret"].get("LoadBalancers"):
            ret = await hub.exec.boto3.client.elbv2.describe_listeners(
                ctx, LoadBalancerArn=load_balancer.get("LoadBalancerArn")
            )
            if not ret["result"]:
                hub.log.debug(
                    f"Could not describe aws.elbv2.listener(s) for aws.elbv2.load_balancer with ARN "
                    f"{load_balancer.get('LoadBalancerArn')}: {ret['comment']}"
                )
                continue
            if ret["ret"].get("Listeners"):
                for listener in ret["ret"]["Listeners"]:
                    tags_ret = await hub.exec.boto3.client.elbv2.describe_tags(
                        ctx, ResourceArns=[listener.get("ListenerArn")]
                    )
                    if not tags_ret["result"]:
                        # Error fetching tags. If fetching tags itself fails, skip and continue. Since tags are
                        # optional, they need not be associated with all listener(s). But API needs to succeed.
                        hub.log.warning(
                            f"Failed listing tags for aws.elbv2.listener '{listener.get('ListenerArn')}'"
                            f"Describe will skip this aws.elbv2.listener and continue. "
                        )
                        continue

                    tags = []
                    if tags_ret.get("ret") and tags_ret.get("ret")["TagDescriptions"]:
                        tags = tags_ret["ret"]["TagDescriptions"][0].get("Tags")

                    certificates_ret = await hub.exec.boto3.client.elbv2.describe_listener_certificates(
                        ctx, ListenerArn=listener.get("ListenerArn")
                    )
                    if not certificates_ret["result"]:
                        # Since certificates are optional, they need not be associated with all listener(s).
                        # But API needs to succeed.
                        hub.log.warning(
                            f"Failed listing certificates for aws.elbv2.listener '{listener.get('ListenerArn')}'"
                            f"Describe will skip this aws.elbv2.listener and continue. "
                        )
                        continue

                    certificates = []
                    if certificates_ret.get("ret") and certificates_ret.get("ret").get(
                        "Certificates"
                    ):
                        certificates = certificates_ret["ret"]["Certificates"]

                    resource_converted = hub.tool.aws.elbv2.conversion_utils.convert_raw_listener_to_present(
                        raw_resource=listener,
                        idem_resource_name=listener.get("ListenerArn"),
                        tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
                        certificates=certificates,
                    )
                    result[resource_converted["resource_id"]] = {
                        "aws.elbv2.listener.present": [
                            {parameter_key: parameter_value}
                            for parameter_key, parameter_value in resource_converted.items()
                        ]
                    }
    return result



async def get_alias_by_name(hub, ctx, alias_name):
    """
    This function will take alias name as input
    and returns the metadata like target kms key, target arn about the alias name
    by searching in aliases.

    Args:
        hub:
        ctx:
        alias_name: alias name of the KMS key

    Returns:
        Dict[str, str]
    """
    ret = {}
    result = dict(comment="", result=True, ret=None)
    try:
        before = await hub.exec.boto3.client.kms.list_aliases(ctx)
        result["result"] = before["result"]
        result["comment"] = before["comment"]
        if before["result"] and before["ret"].get("Aliases"):
            for alias in before["ret"]["Aliases"]:
                if alias_name == alias.get("AliasName"):
                    ret = alias
                    break
        result["ret"] = ret
    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False

    return result



import copy
from typing import Any
from typing import Dict

"""
Util functions for SQS Queue
"""

DEFAULT_ATTRIBUTES_VALUES = {
    "delay_seconds": 0,
    "maximum_message_size": 262144,
    "message_retention_period": 345600,
    "policy": None,
    "receive_message_wait_time_seconds": 0,
    "redrive_policy": None,
    "visibility_timeout": 30,
    "kms_master_key_id": None,
    "kms_data_key_reuse_period_seconds": 300,
    "sqs_managed_sse_enabled": False,
    "fifo_queue": False,
    "content_based_deduplication": False,
    "deduplication_scope": "queue",
    "fifo_throughput_limit": "perQueue",
}


def compare_present_queue_attributes(
    hub, expected_attributes: Dict[str, Any], actual_attributes: Dict[str, Any]
) -> bool:
    """
    Checks if the expected_attributes are contained withing actual_attributes. A None value in expected_attributes and a
        default value in actual_attributes are considered equal

    Args:
        hub: required for functions in hub
        expected_attributes(Dict): the expected attributes
        actual_attributes(Dict): the actual attributes

    Returns:
        True if expected_attributes is contained within actual_attributes, False otherwise
    """

    for (
        expected_attribute_name,
        expected_attribute_value,
    ) in expected_attributes.items():
        actual_attribute_value = actual_attributes.get(expected_attribute_name)

        if actual_attribute_value is None:
            return actual_attribute_value == expected_attribute_value

        if expected_attribute_name == "policy":
            if not hub.tool.aws.sqs.queue_utils.are_aws_policies_equal(
                expected_attribute_value, actual_attribute_value
            ):
                return False
            continue

        if expected_attribute_value != actual_attribute_value:
            if (
                expected_attribute_value is None
                and actual_attributes.get(expected_attribute_name)
                == DEFAULT_ATTRIBUTES_VALUES[expected_attribute_name]
            ):
                continue

            return False

    return True


def are_aws_policies_equal(
    hub, expected_policy: Dict[str, Any], actual_policy: Dict[str, Any]
) -> bool:
    """
    Compares two AWS policies

    Args:
        hub: required for functions in hub
        expected_policy(Dict): the expected policy
        actual_policy(Dict): the actual policy

    Returns:
        True if the policies are equal, False otherwise
    """
    if expected_policy.get("Id") != actual_policy.get("Id") or expected_policy.get(
        "Version"
    ) != actual_policy.get("Version"):
        return False

    expected_statement_list = expected_policy.get("Statement", [])
    actual_statement_list = actual_policy.get("Statement", [])

    if len(expected_statement_list) != len(actual_statement_list):
        return False

    expected_statement_list = [
        hub.tool.aws.sqs.queue_utils.sort_policy_statement_lists(v)
        for v in expected_statement_list
    ]
    actual_statement_list = [
        hub.tool.aws.sqs.queue_utils.sort_policy_statement_lists(v)
        for v in actual_statement_list
    ]

    for expected_statement in expected_statement_list:
        if expected_statement not in actual_statement_list:
            return False

    return True


def sort_policy_statement_lists(
    hub, policy_statement: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Sorts all lists in an AWS policy statement

    Args:
        hub: required for functions in hub
        policy_statement(Dict): the policy statement whose lists will be sorted

    Returns:
        AWS policy statement with sorted lists
    """

    sorted_policy_statement = copy.deepcopy(policy_statement)

    if "Principal" in sorted_policy_statement:
        principal = sorted_policy_statement.get("Principal")
        if isinstance(principal, Dict):
            for key, value in principal.items():
                if isinstance(value, list):
                    sorted_policy_statement["Principal"][key] = sorted(value)
    elif "NotPrincipal" in sorted_policy_statement:
        not_principal = sorted_policy_statement.get("NotPrincipal")
        if isinstance(not_principal, Dict):
            for key, value in not_principal.items():
                if isinstance(value, list):
                    sorted_policy_statement["NotPrincipal"][key] = sorted(value)

    action_block = sorted_policy_statement.get("Action")
    if isinstance(action_block, list):
        sorted_policy_statement["Action"] = sorted(action_block)

    if "Resource" in sorted_policy_statement:
        resource_block = sorted_policy_statement.get("Resource")
        if isinstance(resource_block, list):
            sorted_policy_statement["Resource"] = sorted(resource_block)
    elif "NotResource" in sorted_policy_statement:
        not_resource_block = sorted_policy_statement.get("NotResource")
        if isinstance(not_resource_block, list):
            sorted_policy_statement["NotResource"] = sorted(not_resource_block)

    condition_map = sorted_policy_statement.get("Condition", {})
    for condition_type_string, condition in condition_map.items():
        for condition_key_string, condition_value in condition.items():
            if isinstance(condition_value, list):
                sorted_policy_statement["Condition"][condition_type_string][
                    condition_key_string
                ] = sorted(map(lambda x: str(x), condition_value))

    return sorted_policy_statement



"""Exec module for managing EC2 ami"""
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    filters: List = None,
    tags: Dict[str, str] = None,
) -> Dict:
    """Get EC2 ami from AWS account.

    Get a single ami from AWS. If more than one resource is found, the first resource returned from AWS
    will be used. The function returns None when no resource is found.

    Args:
        name (str):
            The name of the Idem state.

        resource_id (str, Optional):
            Image id of the ami.

        filters (list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_images

        tags (dict, Optional):
            The dict of tags to filter by in the format ``{tag-key: tag-value}`` . For example, to find all resources
            that have a tag with the key "Owner" and the value "TeamA" , specify ``{Owner: TeamA}``

            * Key (str):
                The key name for the tag to be used to filter by.

            * Value(str):
                The value associated with this tag to filter by.
    """

    result = dict(comment=[], ret=None, result=True)

    tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)

    ret = await hub.tool.aws.ec2.ami.search_raw(
        ctx=ctx, resource_id=resource_id, filters=filters, tags=tags
    )

    if not ret["result"]:
        if "InvalidAMIID.Malformed" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.ami", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Images"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.ami", name=name
            )
        )
        return result

    resource = ret["ret"]["Images"][0]
    if len(ret["ret"]["Images"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type="aws.ec2.ami", resource_id=resource_id
            )
        )

    result["ret"] = hub.tool.aws.ec2.conversion_utils.convert_raw_ami_to_present(
        raw_resource=resource
    )

    return result


async def list_(
    hub,
    ctx,
    name: str,
    filters: List = None,
    tags: Dict[str, str] = None,
) -> Dict:
    """List EC2 ami from AWS account.

    Fetch a list of ami from AWS. The function returns empty list when no resource is found.

    Args:
        name(str):
            The name of the Idem state.

        filters (list, Optional):
            One or more filters, A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_images

        tags (dict, Optional):
            The dict of tags to filter by in the format ``{tag-key: tag-value}`` . For example, to find all resources
            that have a tag with the key "Owner" and the value "TeamA" , specify ``{Owner: TeamA}``

            * Key (str):
                The key name for the tag to be used to filter by.

            * Value(str):
                The value associated with this tag to filter by.
    """
    result = dict(comment=[], ret=[], result=True)

    tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)

    ret = await hub.tool.aws.ec2.ami.search_raw(
        ctx=ctx,
        filters=filters,
        tags=tags,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Images"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.ami", name=name
            )
        )
        return result

    for image in ret["ret"]["Images"]:
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_ami_to_present(
                raw_resource=image
            )
        )
    return result



"""Tag helper for AWS Neptune resources."""
import copy
from typing import Any
from typing import Dict


async def get_tags_for_resource(hub, ctx, resource_arn: str = None):
    """Gets the tags for a neptune resource.

    Args:
        resource_arn: aws resource arn

    Returns:
        Dict[str, Any]

    Example:
        {"result": True|False, "comment": "A message tuple", "ret": Dict['str', Any]|None}
    """
    result = dict(comment=(), result=False, ret=None)
    if not resource_arn:
        return result
    tags = await hub.exec.boto3.client.neptune.list_tags_for_resource(
        ctx, ResourceName=resource_arn
    )
    tags_ret = tags.get("ret").get("TagList") if tags.get("result") else None
    result["ret"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags_ret)
    result["result"] = tags.get("result")
    if not result["result"]:
        result["comment"] = (
            f"Getting tags for resource with resource_arn: '{resource_arn}' failed",
        )
    return result


async def update_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """Update tags of AWS Neptune resources.

    Args:
        hub:
        ctx:
        resource_arn: aws resource arn
        old_tags: dict in the format of {tag-key: tag-value}
        new_tags: dict in the format of {tag-key: tag-value}

    Returns:
        Dict[str, Any]

    Example:
        {"result": True|False, "comment": "A message tuple", "ret": dict of updated tags}
    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.neptune.remove_tags_from_resource(
                ctx,
                ResourceName=resource_arn,
                TagKeys=[key for key, value in tags_to_remove.items()],
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.neptune.add_tags_to_resource(
                ctx,
                ResourceName=resource_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

from deepdiff import DeepDiff

"""
 Utility functions to support aws auto scaling group operations
"""


def convert_raw_auto_scaling_group_to_present(
    hub, auto_scaling_group: Dict[str, Any]
) -> Dict[str, Any]:
    result = {}
    if auto_scaling_group is None:
        return result

    resource_parameters = OrderedDict(
        {
            "LaunchConfigurationName": "launch_configuration_name",
            "LaunchTemplate": "launch_template",
            "InstanceId": "instance_id",
            "MinSize": "min_size",
            "MaxSize": "max_size",
            "DesiredCapacity": "desired_capacity",
            "DefaultCooldown": "default_cooldown",
            "AvailabilityZones": "availability_zones",
            "LoadBalancerNames": "load_balancer_names",
            "TargetGroupARNs": "target_group_ar_ns",
            "HealthCheckType": "health_check_type",
            "HealthCheckGracePeriod": "health_check_grace_period",
            "PlacementGroup": "placement_group",
            "MixedInstancesPolicy": "mixed_instances_policy",
            "VPCZoneIdentifier": "vpc_zone_identifier",
            "TerminationPolicies": "termination_policies",
            "NewInstancesProtectedFromScaleIn": "new_instances_protected_from_scale_in",
            "CapacityRebalance": "capacity_rebalance",
            "LifecycleHookSpecificationList": "lifecycle_hook_specification_list",
            "Tags": "tags",
            "MaxInstanceLifetime": "max_instance_lifetime",
            "DesiredCapacityType": "desired_capacity_type",
        }
    )

    tag_parameters = OrderedDict(
        {"Key": "key", "Value": "value", "PropagateAtLaunch": "propagate_at_launch"}
    )

    result = {
        "name": auto_scaling_group.get("AutoScalingGroupName"),
        "resource_id": auto_scaling_group.get("AutoScalingGroupName"),
    }
    for parameter_key, parameter_value in resource_parameters.items():
        if parameter_key in auto_scaling_group:
            result[parameter_value] = auto_scaling_group.get(parameter_key)

    # Special handling for tags
    translated_tags = {}
    for tag in auto_scaling_group.get("Tags"):
        translated_tags[tag["Key"]] = tag["Value"]
        translated_tags[f"propagate_at_launch-{tag['Key']}"] = tag["PropagateAtLaunch"]
    result["tags"] = translated_tags
    return result


def is_auto_scaling_group_updated(
    hub,
    current_state: Dict[str, Any],
    desired_state: Dict[str, Any],
) -> bool:
    if not current_state:
        return True
    list_parameters = [
        "availability_zones",
        "load_balancer_names",
        "target_group_ar_ns",
        "termination_policies",
        "lifecycle_hook_specification_list",
    ]
    dict_parameters = ["launch_template", "mixed_instances_policy"]
    exclusion_list = ["tags"]
    for parameter in desired_state:
        if parameter in exclusion_list:
            continue
        if desired_state.get(parameter) is None:
            # To favour optional parameters of resource
            continue
        if parameter == "vpc_zone_identifier":
            current_zone_ids = (
                current_state.get("vpc_zone_identifier").split(",")
                if current_state.get("vpc_zone_identifier")
                else []
            )
            desired_zone_ids = (
                desired_state.get("vpc_zone_identifier").split(",")
                if desired_state.get("vpc_zone_identifier")
                else []
            )
            if not hub.tool.aws.state_comparison_utils.are_lists_identical(
                current_zone_ids, desired_zone_ids
            ):
                return True
            else:
                continue
        if parameter in list_parameters:
            if not hub.tool.aws.state_comparison_utils.are_lists_identical(
                current_state.get(parameter), desired_state.get(parameter)
            ):
                return True
            else:
                continue
        if parameter in dict_parameters:
            if DeepDiff(
                current_state.get(parameter),
                desired_state.get(parameter),
                ignore_order=True,
            ):
                return True
            else:
                continue
        if current_state.get(parameter) != desired_state.get(parameter):
            return True
    return False


def convert_present_tags_to_raw_tags(
    hub, resource_id, tags: [Dict[str, Any]]
) -> [Dict[str, Any]]:
    result = []
    for tag in tags:
        result.append(convert_tag_to_raw(resource_id, tag))
    return result


def convert_tag_to_raw(resource_id: str, tag: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "ResourceType": "auto-scaling-group",
        "ResourceId": resource_id,
        "Key": tag.get("key"),
        "PropagateAtLaunch": tag.get("propagate_at_launch"),
        "Value": tag.get("value"),
    }
